#!/usr/bin/env python3
#
#   (c) 2012  Universite de Rennes 1
#
# Contact address: <t3devkit@irisa.fr>
#
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
#

#
# CoAP link format messages based on RFC 6690
#

import re

from	ttproto.core.data		import *
from	ttproto.core.typecheck	import *
from	ttproto.core.list		import ListValue
from	ttproto.core.primitive	import IntValue, StrValue
from	ttproto.core			import exceptions
from	ttproto.core.packet		import Optional


#__all__ = [
#	'CoAPLink',
#]

IntValue.build_string = lambda self: (self, str (self))
_re_int = re.compile ("([0-9]+)")
@classmethod
def decode_string_int (cls, s):
	mo = _re_int.match (s)
	if not mo:
		raise Exception ("not an integer")

	return IntValue (mo.group(1)), s[mo.end(1):]
IntValue.decode_string = decode_string_int

StrValue.build_string = lambda self: (self, self)
StrValue.decode_string = classmethod (lambda cls, s: (cls (s), ""))

#TODO: assert is_flat in build_string

class TextListValue (ListValue):
	@classmethod
	@typecheck
	def metaclass_func (cls, name, bases, classdict, content_type: is_type, is_ordered: bool, delimiter: str, prepend: bool, reg: optional (either (tuple, str))):
		result = ListValue.metaclass_func (name, bases, classdict, content_type, is_ordered)

		# bless it
		# TODO: make metaclass_func a static function
		assert len (result.__bases__) == 1
		result.__bases__ = cls,
		result.__cls_delimiter = delimiter
		result.__cls_prepend = prepend
		if isinstance (reg, str):
			result.__cls_reg = re.compile (reg)
		elif isinstance (reg, tuple):
			result.__cls_reg = re.compile (*reg)
		else:
			result.__cls_reg = None

		return result

	def build_string (self):
		if not self:
			return self, ""
		else:
			values, strs = zip (*[v.build_string() for v in self])

			return type (self) (values), (self.__cls_delimiter + self.__cls_delimiter.join (strs))

	@classmethod
	def decode_string (cls, s: str, count = None):

		assert count is None or count >= 0

		values = []
		delimiter = cls.__cls_delimiter
		reg       = cls.__cls_reg
		prepend   = cls.__cls_prepend

		try:
			if count is not None:
				do_it = count > 0
			elif not s:
				do_it = False
			elif prepend:
				do_it = s.startswith (delimiter)
			elif reg:
				do_it = reg.match (s)
			else:
				do_it = True

			if do_it:
				if prepend:
					s = eat (s, delimiter)

				if reg:
					mo = reg.match (s)
					if not mo and prepend:
						raise Exception ("expected regex %r" % mo.pattern)
				while True:
					if reg:
						v, remainder = cls.get_content_type().decode_string (mo.group(0))
						if remainder:
							raise Exception ("unexpected %r" % remainder)

						s = s[mo.end(0):]
					else:
						v, s = cls.get_content_type().decode_string (s)
					values.append (v)

					if count is None:
						if not s or not s.startswith (delimiter):
							break
					else:
						count -= 1
						if not count:
							break

					s = eat (s, delimiter)

					if reg:
						mo = reg.match (s)
						if not mo:
							if count is None and not delimiter:
								break
							raise Exception ("expected regex %r" % mo.pattern)


		except Exception as e:
			exceptions.push_location (e, cls, str (len (values)))
			raise

		return cls (values), s

def TextOrderedListClass (name, bases, classdict, content_type, delimiter = "", prepend = False, reg = None):
	return TextListValue.metaclass_func (name, bases, classdict, content_type, True, delimiter, prepend, reg)

# TODO: merge it with Packet
# TODO: replace Packet.__init__(*k) with Packet.__init__(k)
# TODO: make TextRecord(s) an alias to TextRecord.decode_string(s)
class TextRecordValue (Value):
	class Tag:
		def __init__ (self, default: optional (Value) = None):
			"""Initialise the tag

			'default' is the default value to be used when the
			field is left undefined.
			"""
			self.__default = default

		def __repr__ (self):
			return "%s(%s)" % (type (self).__name__, repr (self.__default))

		def get_default_value (self):
			"""Return the default value for this field, or None if there is no default value"""
			return self.__default

		def build_string (self, value, ctx = None):
			return value.build_string()

		def decode_string (self, type_, s, ctx = None):
			return type_.decode_string (s)

		@typecheck
		def init (self, parent: type, field_id: int, field):
			assert parent.get_field (field_id) is field
			assert field.tag is self

			self.__parent_type = parent
			self.__field_id = field_id
			self.__field = field
			self.__default = self.__field.store_data (self.__default, none_is_allowed = True)

		def get_parent_type (self):
			return self.__parent_type

		def get_field (self):
			return self.__field

		def get_field_id (self):
			return self.__field_id

	class Field:
		@typecheck
		def __init__ (self, name: str, alias: str, typ: either (is_type, Optional), tag = None):

			self.name = name
			self.alias = alias
			self.optional = isinstance (typ, Optional)
			self.type = get_type (typ.type if self.optional else typ)
			if isinstance (tag, TextRecordValue.Tag):
				self.tag = tag
			else:
				self.tag = TextRecordValue.Tag (self.store_data (tag))

		def __repr__ (self):
			fmt_type = "Optional(%s)" if self.optional else "%s"
			return "TextRecordValue.Field(%s, %s, %s, %s)" % (
				repr (self.name),
				repr (self.alias),
				fmt_type % repr (self.type),
				repr (self.tag)
			)

		@typecheck
		def store_data (self, data, none_is_allowed = True):
			"""shortcut to store_data() using the Field's details (type, optional)"""

			return store_data (data, self.type, none_is_allowed = none_is_allowed, omit_is_allowed = self.optional)

		def decode_string (self, s, ctx = None):
			return self.tag.decode_string (self.type, s, ctx)

	@classmethod
	@typecheck
	def __init_fields (cls, fields: optional (list_of (tuple))):
		# check the validity of the fields
		if fields is not None:
			for f in fields:
				assert 3 <= len(f) <= 4
				assert type (f[0]) == str
				assert type (f[1]) == str
				assert is_type (f[2]) or isinstance (f[2], Optional)

		cls.__cls_fields = []

		if fields:
			for f in fields:
				# ensure there is no name/alias collision in the field names
				for previous_field in cls.__cls_fields:
					for name in f[0:2]:
						assert name != previous_field.name
						assert name != previous_field.alias

				cls.__cls_fields.append (TextRecordValue.Field (*f))

		# initialise the tags in the new fields
		i = 0
		for f in cls.__cls_fields:
			f.tag.init (cls, i, f)
			i+=1

	@classmethod
	def metaclass_func (cls, name, bases, classdict,
				fields:	list_of (tuple),
				format: optional (str) = None
				):
		assert len (bases) == 0 # FIXME: it could be useful to have multiple inheritance in some cases ?


		result = type (name, (cls,), classdict)

		result.__init_fields (fields)
		result.__cls_format = format

		if format is not None:
			assert format.count ("#") % 2 == 0

			elems = []
			is_field = False
			for e in format.split ("#"):
				if is_field:
					is_field = False
					if e:
						# field reference
						elems.append (result.get_field_id (e))
						is_field = False
						continue
					else:
						# "##" escape string
						e = "#"
				else:
					is_field = True

				if not e:
					continue

				# append e to the elems
				if elems:
					# stick it to the previous one (if any)
					if isinstance (elems[-1], str):
						elems[-1] += e
						continue

				elems.append (e)
			result.__cls_elems = elems

		return result

	@classmethod
	@typecheck
	def get_field (cls, id: either (int, str)) -> Field:
		"""Access one individual Field either by name, alias or numerical index

		Raise exceptions.UnknownField in case the requested field does not exist
		"""
		if isinstance (id, str):
			id = cls.get_field_id (id)
		return cls.__cls_fields[id]

	@classmethod
	@typecheck
	def get_field_id (cls, arg: either(str, int)) -> int:
		"""Get the numerical index of a field, given its name or numerical index

		Raise exceptions.UnknownField in case the requested field does not exist
		"""
		if isinstance (arg, int):
			assert -cls.get_length() <= arg < cls.get_length()
			return arg % cls.get_length()
		else:
			i=0
			for f in cls.fields():
				if f.name == arg or f.alias == arg:
					return i
				i+=1

		raise exceptions.UnknownField("Unknown field name", arg)

	@typecheck
	def __getitem__ (self, index: either (int, str)) -> optional (is_data):
		"""Access one field by id or name or by type (shortcut for find_type())

		The following indexing ways are supported:
		- positive or null integer
		- negative integer
		- field name
		- field alias
		"""
		if isinstance (index, str):
			index = self.get_field_id (index)
		else:
			# int
			assert -len (self) <= index < len (self)

			if index < 0:
				index += len (self)

		return self.__datas[index]

	@typecheck
	def __init__ (self, *k, **kw):

		super().__init__()

		# TODO: copy constructor ?

		if len(k):
			if len (kw):
				raise Error ('cannot mix positional and named arguments')

			missing = len (self) - len (k)
			if missing < 0:
				raise Error ("to many parameters")

			self.__datas = [f.store_data (d) for d,f in zip (k, self.fields())]

			if missing:
				self.__datas.extend([None for v in range (0, missing)])
		else:
			# named parameters

			self.__datas = []
			for field in self.fields():
				for n in field.name, field.alias:
					if n in kw:
						v = kw.pop (n)
						break
				else:
					v = None

				# append the type to the internal list, but possibly convert it before
				self.__datas.append (field.store_data (v))

			if len (kw):
				raise Error ('unknown fields in type %s: %s' % (type(self).__name__, ' '.join (kw.keys())))

	@classmethod
	def get_length (cls):
		return len (cls.__cls_fields)

	@classmethod
	@typecheck
	def fields (cls) -> iterable:
		return iter (cls.__cls_fields)

	__len__ = get_length

	def _repr (self):
		result = []
		for field, data in zip (self.__cls_fields, self.__datas):
			if data is not None:
				result.append ("%s=%s" % (field.alias, repr (data)))
		return "%s(%s)" % (type (self).__name__, ", ".join (result))

	def build_string (self):
		#TODO assert is flat
		#TODO support tags
		#TODO what to do with None fields

		values = []
		strs = []
		for d in self.__datas:
			if d is None:
				v, s = None, None
			else:
				v, s = d.build_string()
			values.append (v)
			strs.append (s)

		return type (self) (*values), "".join (
			(e if isinstance (e, str) else strs[e])  for e in self.__cls_elems
		)
	@classmethod
	def decode_string (cls, s):
		#TODO support tags

		values = [None]*cls.get_length()

		for e in cls.__cls_elems:
			if isinstance (e, str):
				if not s.startswith (e):
					raise Exception ("Expected %r" % e)
				s = s[len(e):]
			else:
				values[e], s = cls.get_field(e).decode_string(s)
		#TODO: exception location
		return cls (*values), s


	def _is_flat (self):
		#TODO
		return True

	@typecheck
	def get_datas (self, index: either (int, str)) -> iterable:
		seq = self
		index = self.get_field_id (index)
		while seq:
			v = seq.__datas [index]
			if v is not None:
				yield v
			seq = seq.get_parent()


	def _display (self, indent, output):
		indent +=2
		import ttproto.core.list

		print("###[ %s ]###" % type(self).__name__, file = output)

		i=0
		for f in self.fields():
			hdr = "%s%s= " % (' ' * indent, f.name)

			#FIXME: ugly hack -> do something more generic
			if issubclass (f.type, core.list.ListValue):
				print (hdr,file=output)
				for v in self.get_datas (i):
					v._display (indent, output)
			else:
				datas = []
				for d in self.get_datas (i):
					s = str (d)
					datas.append (s)

				print("%s%s%s" % (
					hdr,
					" " * (0 if len (hdr) >= 28 else 28-len(hdr)),
					" & ".join(datas),
				), file = output)
			i+=1

TextRecordClass = TextRecordValue.metaclass_func

class Reg (TextRecordValue.Tag):
	def __init__ (self, pattern, flags = 0):
		super().__init__()
		self.__pattern = pattern
		self.__re = re.compile (pattern, flags)

	# TODO: validate the strings we send

	def decode_string (self, type_, s, ctx = None):
		mo = self.__re.match (s)
		if not mo:
			raise Exception ("expected string matching regex %r" % self.__pattern)

		# number of groups in the regex
		n = len (mo.regs) - 1

		assert n <= 2 # must not have more than 2 () groups

		v, remainder = type_.decode_string (mo.group (n))
		if remainder:
			raise Exception ("unexpected: %r" % remainder)

		s = s[mo.end (1 if n == 2 else 0):]

		return v, s

_SCHEME = "([A-Z](?:[A-Z][0-9])*):"

def eat (s, token):
	if s.startswith (token):
		return s[len(token):]
	else:
		raise Exception ("expected %r" % token)

class CoAPLinkParam (
	metaclass = TextRecordClass,
	fields    = [
		("Name",	"name",		str),
		("Value",	"val",		str),
	]):

	__quoted = "anchor", "title"
	__unquoted = "hrefland sz title*".split()

	__re_tokens = re.compile(r"[]!#$%&'()*+./0-9:<=>?@a-z[^_`{|}~-]+", re.I)
	__re_parmname = re.compile (r"([a-zA-Z0-9!#$&+.^_`|~-]+|title\*)=")

	@classmethod
	def str_must_be_quoted (cls, s):
		mo = cls.__re_tokens.match (s)
		return not mo or mo.end(0) != len (s)

	@classmethod
	def param_must_be_quoted (cls, name):
		return name in cls.__quoted

	@classmethod
	def param_must_not_be_quoted (cls, name):
		return name in cls.__unquoted

	def build_string (self):
		name = self["name"]
		val = self["val"]

		if self.param_must_be_quoted (name):
			quote = True

		elif self.str_must_be_quoted (val):
			quote = True

			if self.param_must_not_be_quoted (name):
				raise Exception ("Param value contains invalid tokens (needs quoting)")

		else:
			quote = False

		if quote:
			val = '"%s"' % val.replace ('"', '\\"').replace ("\\", "\\\\")

		return self, "%s=%s" % (name, val)

	@classmethod
	def decode_string (cls, s):
		mo = cls.__re_parmname.match (s)
		if not mo:
			raise Exception ("invalid parameter name")

		name = mo.group(1)
		s = s[mo.end(0):]

		quoted = s and s[0] == '"'

		if quoted:
			if cls.param_must_not_be_quoted (name):
				raise Exception ("parameter %s must not be quoted" % name)

			val = []
			i = 1
			quote = False
			while i < len(s):
				c = s[i]
				if quote:
					quote = False
					if c == '\"' or c == '\\':
						val.append (c)
					else:
						# was an unquoted \
						val.append ("\\" + c)
				else:
					if c == '\\':
						quote = True
					elif c == '"':
						# end of string
						break
					else:
						val.append (c)
				i += 1
			else:
				raise Exception ("malformatted quoted string")

			val = "".join(val)
			s = s[i+1:]
		else:
			if cls.param_must_be_quoted (name):
				raise Exception ("parameter %s must be quoted" % name)

			mo = cls.__re_tokens.match (s)
			if not mo:
				raise Exception ("malformatted parameter value")

			val = mo.group(0)
			s = s[mo.end(0):]

		return cls (name, val), s







class CoAPLinkParamList (
	metaclass = TextOrderedListClass,
	content_type = CoAPLinkParam,
	delimiter = ";",
	prepend = True,
	reg = "[^;,]*",
):
	pass

class CoAPLinkValue (
	metaclass = TextRecordClass,
	fields    = [
		("UriReference",	"uri",	str, 	Reg("[^>]*")),
		("params",		"par",	CoAPLinkParamList),
	],
	format    = "<#uri#>#par#",
	):
	pass

class CoAPLink (
	metaclass = TextOrderedListClass,
	content_type = CoAPLinkValue,
	delimiter = ",",
):

	@classmethod
	def _decode_message (cls, bin_slice):

		if bin_slice.get_bit_length() % 8:
			raise Exception("Textual message length must be a multiple of 8 bits")

		s = str (bin_slice.raw(), "utf-8")

		v, s = cls.decode_string (s)

		if s:
			raise Exception ("String not fully decoded (%d bytes) left" % len(s))

		return v, bin_slice.shift_bits (bin_slice.get_bit_length())

	@typecheck
	def _build_message (self: is_flat_value) -> is_flatvalue_binary:

		v, s = self.build_string()

		return v, bytes (s, "utf-8")




from ttproto.tat_coap.testcases.td_coap_core_22 import TD_COAP_CORE_22


class TD_COAP_CORE_27 (TD_COAP_CORE_22):
    """
---
TD_COAP_CORE_27:
    cfg: CoAP_CFG_03
    obj: Perform GET transaction with responses containing the ETag option and
         requests containing the If-Match option (CON mode)
    pre:
        - Proxy is configured as a reverse-proxy for the server
        - Proxy’s cache is cleared
        - Server should offer a /validate resource with resource content is not
          empty
        - Client & server supports ETag option and If-Match option
    ref: '[1] clause 5.8.1, 5.10.7,5.10.9,12.1.12, 8.2.2,8.2.1,10.2.2,11.2'
    seq:
        -   s: Client is requested to send a confirmable GET request to proxy
        -   c: Proxy forwards the request to server
        -   c:
            - 'Forwarded request must contain:'
            -   - Type = 0 (CON)
                - Code = 1 (GET)
        -   c:
            - 'Server sends response containing:'
            -   - Code = 69 (2.05 content)
                - Option type = ETag
                - Option value = an arbitrary ETag value
                - Not empty payload
        -   c: Proxy forwards the response to client
        -   s: Client is requested to send a confirmable PUT request to Proxy
        -   c:
            - 'Sent request must contain:'
            -   - Type = 0 (CON)
                - Code = 3 (PUT)
                - Option Type=If-Match
                - Option value=ETag value received in step 4
                - An arbitrary payload (which differs from the payload received
                  in step 3)
        -   v: Proxy forwards the request to servers resource & server updates
               the resource
        -   c:
            - 'Server sends response containing:'
            -   - Code = 68 (2.04 Changed)
                - Option type = ETag
                - Option value = an arbitrary ETag value which differs from the
                  ETag received in step 4
        -   c: Proxy forwards the response to client
        -   c:
            - 'Forwarded response contains:'
            -   - Code = 68 (2.04 Changed)
                - Option type = ETag
                - Option value = same ETag value found in step 8
        -   v: Client displays the response
        -   s: Update the content of the server’s resource from a CoAP client
        -   s: Client is requested to send s confirmable PUT request to proxy
               so as to perform an atomic update
        -   c:
            - 'Sent request must contain:'
            -   - Type = 0 (CON)
                - Code = 3 (PUT)
                - Option Type=If-Match
                - Option value=ETag value received in step 8
                - An arbitrary payload (which differs from the previous
                  payloads)
        -   c: Proxy forwards the request to server’s resource
        -   c:
            - 'Sent request must contain:'
            -   - Type = 0 (CON)
                - Code = 3 (PUT)
                - Option Type=If-Match
                - Option value=same ETag value found in step 14
                - An arbitrary payload (which differs from the previous
                  payloads)
        -   c:
            - 'Server sends response containing:'
            -   - Code = 140 (4.12 Precondition Failed)
        -   v: Proxy forwards the response to client
        -   c:
            - 'Response contains:'
            -   - Code = 140 (4.12 Precondition Failed)
        -   v: Client displays the response
    """

    reverse_proxy = True



#!/usr/bin/env python3
#
#   (c) 2012  Universite de Rennes 1
#
# Contact address: <t3devkit@irisa.fr>
#
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

# FIXME: this is a very dirty script

from ttproto.core import clock, port
from ttproto.core.typecheck import *

from . import local_config
from . import implem
from . import implembR
from . import implemR
from .common import *

import socket, threading
import ttproto.core.lib.inet.all

class UdpServerPort (port.RawMessagePort):

	@typecheck
	def __init__ (self, port_number: int, decode_type: is_type = bytes, endpoint: optional (port.BaseMessagePort) = None):
		super().__init__ (decode_type, endpoint)

		self.__sock = socket.socket (socket.AF_INET6, socket.SOCK_DGRAM)
		self.__sock.bind (("", port_number))

		self.__lock = threading.Lock()
		self.__kill_event = threading.Event()
		self.__peer = None

		self.__thread = threading.Thread (target = self.__thread_func)
		self.__thread.daemon = True
		self.__thread.start()

	def __thread_func (self):
		try:
			while True:
				buff, peer = self.__sock.recvfrom (65535)
				if self.__kill_event.is_set():
					# the port is being killed -> just return silently
					return
				with self.__lock:
					if self.__peer != peer:
						print ("Using new endpoint:", peer)
						self.__peer = peer

				# forward it
				self._forward (buff)
				#TODO: warning if there is no endpoint ?

		except Exception as e:
			print("Error in receiving thread")
			traceback.print_exc()

			#TODO: report the error to the tester

	def enqueue (self, msg: Message):
		assert isinstance (msg.get_binary(), bytes) # supports only byte-aligned data

		with self.__lock:
			peer = self.__peer

		if peer:
			self.__sock.sendto (msg.get_binary(), peer)
		else:	
			print ("%s: warning: ignored (no known peer yet)" % type(self).__name__)
@testcase
def dummy_implem():
	pass


emss_orig = port.EventMessageSent.summary
emrs_orig = port.EventMessageReceived.summary
def emss (self):
	self[1].display()
	return emss_orig (self)
def emrs (self):
	self[1].display()
	return emrs_orig (self)
port.EventMessageSent.summary = emss
port.EventMessageReceived.summary = emrs

with LoggerGroup ([ConsoleLogger(), HTMLLogger()]) as logger:

	logger.log_event (EventTestcaseStarted (dummy_implem))

	imp = implem.SixLowpanImplementation (compression = "-c" in sys.argv)
	imp.get_message_port().add_logger (logger)
	udp = UdpServerPort (13000, Ieee802154, imp.get_message_port())

	imp.daemon = True
	imp.start()
	while True:
		time.sleep(86400)




#!/usr/bin/env python3

from ..common import *


class TD_COAP_CORE_09 (CoAPTestCase):
    """
---
TD_COAP_CORE_09:
    cfg: CoAP_CFG_BASIC
    not: Steps 3 and 4 may occur out-of-order
    obj: Perform GET transaction with separate response (CON mode, no
        piggyback)
    pre: Server offers a resource /separate which is not served immediately
        and which therefore is not acknowledged in a piggybacked way.
    ref: '[COAP] 5.8.1, 5.2.2'
    seq:
    -   s: "Client is requested to send a confirmable GET request to
            server\u2019s resource"
    -   c:
        - 'The request sent by the client contains:'
        -   - Type = 0 (CON)
            - Code = 1 (GET)
            - "Client-generated Message ID (\u2794 CMID)"
            - Uri-Path option "separate"
            - "Client-generated Token (\u2794 CTOK)"
    -   c:
        - 'Server sends response containing:'
        -   - Type = 2 (ACK)
            - Code = 0
            - Message ID = CMID
            - Empty Payload
    -   n: Some time (a couple of seconds) elapses.
    -   c:
        - 'Server sends response containing:'
        -   - Type = 0 (CON)
            - Code = 2.05 (Content)
            - "Server-generated Message ID (\u2794 SMID)"
            - Token = CTOK
            - Content-format option
            - Non-empty Payload
    -   c:
        - 'Client sends response containing:'
        -   - Type = 2 (ACK)
            - Code = 0
            - Message ID = SMID
            - Empty Payload
    -   v: Client displays the response
    """

    @classmethod
    @typecheck
    def get_stimulis(cls) -> list_of(Value):
        """
        Get the stimulis of this test case. This has to be be implemented into
        each test cases class.

        :return: The stimulis of this TC
        :rtype: [Value]
        """
        return [
            CoAP(type='con', code='get')
        ]

    def run(self):
        self.match ("client", CoAP (type="con", code = "get",
                        opt = self.uri ("/separate")))
        CMID = self.coap["mid"]
        CTOK = self.coap["tok"]

        self.next()

        #FIXME: may be out-of-order
        if not self.match("server", CoAP (type="ack", code=0, mid=CMID, pl = b"")):
            raise self.Stop()

        self.next()

         # FIXME: this is in a different conversation
        self.match ("server", CoAP (type="con", code=2.05))
        self.match ("server", CoAP (
                        tok=CTOK,
                        pl = Not (b''),
                        opt= Opt(CoAPOptionContentFormat())
                ), "fail")
        SMID = self.coap["mid"]

        self.next()

        self.match ("client", CoAP (type="ack", code=0, mid=SMID,pl=b""))





#!/usr/bin/env python3
#
#   (c) 2012  Universite de Rennes 1
#
# Contact address: <t3devkit@irisa.fr>
#
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import os, sys, threading, time, traceback, types

from ttproto.core.typecheck import *
from ttproto.core import xmlgen, control, logger, port, data
from ttproto.core.list import ListValue
from ttproto.core.packet import PacketValue


__all__ = [
    "HTMLLogger",
]

# TODO: should be configurable
ROOT_DIR = "reports"


def _gen_dir(root_dir):
    assert os.path.exists(root_dir)

    base = time.strftime("%Y-%m-%d--%Hh%M")
    d = base
    n = 1
    while os.path.exists(os.path.join(root_dir, d)):
        n += 1
        d = "%s-%d" % (base, n)

    return d



class HTMLLogger(logger.Logger):
    __verdict_color_map = {
        None: "#FF9090",
        "none": "#FFFFA0",
        "pass": "#B0FFB0",
        "inconclusive": "#DCBAFF",
    }

    class __RuntimeError(Exception):
        def __init__(self, exc_type, exc_value, traceback):
            Exception.__init__(self)
            self.__info = exc_type, exc_value, traceback

        def exc_info(self):
            return self.__info

    class __Exit(Exception):
        pass

    def __init__(self, directory=None):
        self.__lock = threading.Lock()
        self.__usage_count = 0
        self.__requested_dir = directory
        self.__event = None

    def __enter__(self):
        with self.__lock:
            if self.__usage_count == 0:
                if not os.path.exists(ROOT_DIR):
                    os.mkdir(ROOT_DIR)

                session_dir = self.__requested_dir if self.__requested_dir else _gen_dir(ROOT_DIR)
                self.__dir = os.path.join(ROOT_DIR, session_dir)
                os.mkdir(self.__dir)  # the output directory must not exist

                if not self.__requested_dir:
                    # create a symlink current
                    current = os.path.join(ROOT_DIR, "current")
                    if os.path.exists(current):
                        if os.path.islink(current):
                            os.unlink(current)
                            os.symlink(session_dir, current)
                        else:
                            print("Warning: cannot create symlink 'current'")
                    else:
                        os.symlink(session_dir, current)

                self.__tc_count = 0

                self.__generators = self.gen_index(), self.gen_tc_report(), self.gen_tc_messages(), self.gen_tc_mismatch()
                self.__iteration()

            self.__usage_count += 1

        return self

    def __exit__(self, a, b, c):
        with self.__lock:
            if a:
                err = self.__RuntimeError(a, b, c)
                for g in self.__generators:
                    try:
                        g.throw(err)
                    except self.__RuntimeError as e:
                        print("HTML generator %s did no handle exception %s" % (str(g), str(e)))

            self.__usage_count -= 1
            if self.__usage_count == 0:
                for g in self.__generators:
                    g.close()

    def __iteration(self):
        assert self.__lock.locked()

        for g in self.__generators:
            # TODO: catch possible exceptions ?
            try:
                next(g)
            except StopIteration:
                pass

        if type(self.__event) == control.EventTestcaseTerminated:
            self.__tc_count += 1

    @classmethod
    def __verdict_col(cls, v):
        vcm = cls.__verdict_color_map
        return vcm[v] if v in vcm else vcm[None]

    @typecheck
    def log_event(self, event: logger.LogEvent):
        with self.__lock:
            if self.__usage_count:
                self.__event = event
                self.__iteration()

                # clear the event so as not to interfere with the garbage collector
                self.__event = None
            else:
                print("Warning: ignored event:", event, file=sys.stderr)

    def gen_index(self):
        with xmlgen.XHTML10Generator(output=os.path.join(self.__dir, "index.html"), indented=True) as g:
            gctrl = xmlgen.XMLGeneratorControl(g)

            while type(self.__event) != control.EventTestSessionStarted:
                yield

            title = "Test Session Report - " + self.__event.session.get_description()

            g.head.title(title)
            g.h1(title)

            try:
                with g.table(border=1, cellspacing=0, cellpadding=5):
                    with g.tr:
                        g.th("Testcase")
                        g.th("Report")
                        g.th("Messages")
                        g.th("Verdict")

                        while True:
                            while type(self.__event) != control.EventTestcaseStarted:
                                yield
                                if type(self.__event) == control.EventTestSessionTerminated:
                                    raise self.__Exit()

                            with g.tr:
                                g.td(self.__event.testcase.__name__)
                                g.td(align="center").a(href="tc-%03d-report.html" % self.__tc_count)("X")
                                g.td(align="center").a(href="tc-%03d-messages.html" % self.__tc_count)("X")
                                gctrl.flush()

                                while type(self.__event) != control.EventTestcaseTerminated:
                                    yield

                                v = self.__event.verdict
                                g.td(bgcolor=self.__verdict_col(v))(v)

            except GeneratorExit:
                g.b("Warning: the HTML logging module shut down before the end of the test session")

            except self.__RuntimeError as e:
                g.b("Error in the logging module")
                g.pre("".join(traceback.format_exception(*e.exc_info())))

            except self.__Exit:
                g.h1("Summary")
                with g.table(border=1, cellspacing=0, cellpadding=5):
                    total = self.__event.verdict_summary.count()
                    for name, nb in self.__event.verdict_summary:
                        tr = g.tr
                        if nb:
                            tr(bgcolor=self.__verdict_col(name))
                        with tr:
                            g.td(name)
                            g.td(" %d" % nb, align="right")
                            g.td(" %3d%%" % (nb * 100 // total), align="right")

            except Exception:
                print("(blah)")
                g.b("Error in the logging module")
                g.pre("".join(traceback.format_exception(*sys.exc_info())))

        # TODO: return an error status to the logger object
        while True:
            yield

    def gen_tc_report(self):
        while True:

            while type(self.__event) != control.EventTestcaseStarted:
                yield

            with xmlgen.XHTML10Generator(output=os.path.join(self.__dir, "tc-%03d-report.html" % self.__tc_count),
                                         indented=False) as g:
                gctrl = xmlgen.XMLGeneratorControl(g)

                title = "Test Case Report - " + self.__event.testcase.__name__

                g.head.title(title)
                g.h1(title)

                try:
                    with g.table(border=1, cellspacing=0, celpadding=5):
                        with g.tr:
                            g.th("Timestamp")
                            g.th("Info")
                            g.th("Message")
                        while True:
                            e = self.__event

                            with g.tr:
                                g.td(e.get_timestamp())

                                if type(e) in (port.EventMessageSent, port.EventMessageReceived):
                                    if type(e) == port.EventMessageSent:
                                        g.td("Message sent to " + e.port.__name__)
                                    else:
                                        g.td("Message received from " + e.port.__name__)
                                    g.td.a(href="tc-%03d-messages.html#%x" % (self.__tc_count, id(e.message))).tt(
                                        style="white-space: pre")(e.message.get_description())

                                else:
                                    if type(e) == port.EventMessageMismatch:
                                        g.td.a(href="tc-%03d-mismatch.html#%x" % (self.__tc_count, id(e)))(e.summary())
                                    elif type(e) == control.EventTestcaseRuntimeError:
                                        with g.td:
                                            g(e.summary())
                                            g.pre(e.traceback)
                                    elif type(e) == control.EventSetVerdict:
                                        g.td(e.summary(), bgcolor=self.__verdict_col(e.verdict))
                                    elif type(e) == logger.EventStep:
                                        g.td(e.summary(), bgcolor="#e8e8e8")
                                    else:
                                        g.td(e.summary())

                                    g.td(
                                        "")  # g.td  FIXME: xmlgen <td/> not flushed in the right placf without function call

                                if type(e) == control.EventTestcaseTerminated:
                                    raise self.__Exit()

                            yield

                except self.__Exit:
                    pass

        # TODO: return an error status to the logger object
        while True:
            yield

    def gen_tc_messages(self):
        while True:

            while type(self.__event) != control.EventTestcaseStarted:
                yield

            with xmlgen.XHTML10Generator(output=os.path.join(self.__dir, "tc-%03d-messages.html" % self.__tc_count),
                                         indented=True) as g:
                gctrl = xmlgen.XMLGeneratorControl(g)

                title = "Test Case Messages Report - " + self.__event.testcase.__name__

                g.head.title(title)
                g.h1(title)

                nb = 0
                try:
                    while True:
                        e = None
                        while type(e) not in (port.EventMessageSent, port.EventMessageReceived):
                            yield
                            e = self.__event
                            if type(e) == control.EventTestcaseTerminated:
                                raise self.__Exit()

                        nb += 1

                        g.a(name="%x" % id(e.message))
                        g.hr
                        g.p.b("Message #%d %s on %s at %s (%f)" % (
                            nb, "sent" if type(e) == port.EventMessageSent else "received",
                            e.port.__name__,
                            time.ctime(e.get_timestamp()), e.get_timestamp()))
                        with g.p:
                            m = e.message
                            g.pre(str(m.get_description()) + "\n\n")

                            self.display_value(g, m.get_value())

                            with g.pre:
                                g("\n\n")
                                b = m.get_binary()
                                for offset in range(0, len(b), 16):
                                    values = ["%02x" % v for v in b[offset:offset + 16]]
                                    if len(values) > 8:
                                        values.insert(8, " ")

                                    g("%04x   %s" % (offset, " ".join(values)))


                except self.__Exit:
                    pass

        # TODO: return an error status to the logger object
        while True:
            yield

    class __ExtraColumnMatchResult:
        def __init__(self, value, pattern, mismatches: tuple_of(data.Mismatch)):
            self.__iter_diff = iter(mismatches)
            self.__next_diff()

            self.__iter_value_pattern = self.__value_pattern_generator(value, pattern)
            self.__next_value_pattern()

        def __next_value_pattern(self):
            try:
                result = next(self.__iter_value_pattern)
                self.__value, self.__pattern = result
            except StopIteration:
                self.__value, self.__pattern = None, None

            return self.__value, self.__pattern

        def __value_pattern_generator(self, value, pattern):

            yield value, pattern

            # do not go further if pattern is a template, or if its type is different
            if (not isinstance(pattern, data.Value)) or value.get_type() != pattern.get_type():
                return

            # from here we have the same type
            if isinstance(value, PacketValue):
                if value.get_variant() != pattern.get_variant():
                    # variant mismatch -> we only report the first fields
                    for f1, f2, v, p in zip(value.get_variant().fields(), pattern.get_variant().fields(), value,
                                            pattern):
                        if f1 != f2:
                            break

                        for v_p in self.__value_pattern_generator(v, p):
                            yield v_p
                else:
                    for v, p in zip(value, pattern):
                        for v_p in self.__value_pattern_generator(v, p):
                            yield v_p

            elif isinstance(value, ListValue) and value.get_type().is_ordered():
                # TODO: not ordered case
                for v, p in zip(value, pattern):
                    for v_p in self.__value_pattern_generator(v, p):
                        yield v_p

        def __next_diff(self):
            try:
                # NOTE: may have several times the same value, but different patterns (inherited templates)
                self.__diff = next(self.__iter_diff)
            # print ("next diff:", self.__diff.value, self.__diff.pattern)
            except StopIteration:
                self.__diff = None
            # print ("next diff: none")

            return self.__diff

        def __call__(self, g, value):
            d = self.__diff

            if value is not self.__value:
                pattern = None
            else:
                pattern = self.__pattern
                self.__next_value_pattern()

            # print ("iteration:", value, pattern)

            with g.td(
                    bgcolor="#FF9090" if d and d.value is value else "white"):  # FIXME: will not work in case we have strange matches (especially with one template containing values or with an unordered set). the mismatchlist should be a tree

                # ensure that pattern 'contains' d.pattern (possibly multiples d.pattern instances)
                if d and d.value is value:
                    while True:
                        if not d or not d.value is value:
                            break
                        g("%s: expected %s" % (type(d).__name__, d.describe_expected()))
                        g.br

                        d = self.__next_diff()

                elif pattern:
                    if not isinstance(pattern, data.Value):
                        expected = str(pattern)
                    elif pattern.get_type() != value.get_type():
                        if isinstance(pattern, PacketValue):
                            expected = pattern.get_variant().__name__
                        else:
                            expected = pattern.get_type().__name__
                    elif isinstance(value, PacketValue) and pattern.get_variant() != value.get_variant():
                        expected = pattern.get_variant().__name__
                    else:
                        if isinstance(pattern, PacketValue) or isinstance(pattern, ListValue):
                            # just display the type
                            if isinstance(pattern, PacketValue):
                                expected = pattern.get_variant().__name__
                            else:
                                expected = pattern.get_type().__name__
                        else:
                            expected = str(pattern)

                    g("expected %s" % expected)

    def gen_tc_mismatch(self):
        while True:

            while type(self.__event) != control.EventTestcaseStarted:
                yield

            with xmlgen.XHTML10Generator(output=os.path.join(self.__dir, "tc-%03d-mismatch.html" % self.__tc_count),
                                         indented=True) as g:
                gctrl = xmlgen.XMLGeneratorControl(g)

                title = "Mismatch Reports - " + self.__event.testcase.__name__

                g.head.title(title)
                g.h1(title)

                nb = 0
                try:
                    while True:
                        e = None
                        while type(e) != port.EventMessageMismatch:
                            yield
                            e = self.__event
                            if type(e) == control.EventTestcaseTerminated:
                                raise self.__Exit()

                        nb += 1

                        g.a(name="%x" % id(e))
                        g.hr
                        g.p.b("Message mismatch #%d on %s at %s (%f)" % (
                            nb, e.port.__name__,
                            time.ctime(e.get_timestamp()), e.get_timestamp()))
                        with g.p.pre:
                            g(e.message.get_description())
                            g("\n")

                        self.display_value(g, e.message.get_value(),
                                           self.__ExtraColumnMatchResult(e.message.get_value(), e.pattern,
                                                                         e.mismatches))

                except self.__Exit:
                    pass

        # TODO: return an error status to the logger object
        while True:
            yield

    @typecheck
    def calc_output_size(self, value: data.Value):
        if isinstance(value, PacketValue):
            c_lines = 0
            c_cols = 1
            pid = value.get_variant().get_payload_id()
            for i in range(0, len(value)):
                l, c = self.calc_output_size(value[i])
                if i == pid:
                    c -= 1
                c_lines += l
                if c > c_cols:
                    c_cols = c
            result = (1 + c_lines), (2 + c_cols)

        elif isinstance(value, ListValue):
            c_lines = 0
            c_cols = 1
            for i in range(0, len(value)):
                l, c = self.calc_output_size(value[i])
                c_lines += l
                if c > c_cols:
                    c_cols = c
            result = (1 + c_lines), (1 + c_cols)

        else:
            result = (1, 1)

        value.__output_size = result
        return result

    @typecheck
    def display_value(self, g: xmlgen.XMLGenerator, value: data.Value, col_add: optional(callable) = None):
        l, c = self.calc_output_size(value)
        with g.table(border=1, cellspacing=0, cellpadding=2, width=("800px" if col_add else "640px")):
            self.__display_value_internal(g, value, col_add)



    def __display_value_internal(self, g: xmlgen.XMLGenerator, value: data.Value, col_add: optional(callable),
                                 prefix=None, data_span=1, description=None):

        l, c = value.__output_size
        col_style = "border-left-width: medium; border-top-width: medium; border-right: none"
        headline_style = "border-top-width: medium; border-left: none; padding-top: 4pt; padding-bottom: 5pt"
        prefix_style = "padding-left: 10pt; border-left: none"
        if l != 1:
            rowspan = {"rowspan": l}
        else:
            rowspan = {}

        if isinstance(value, PacketValue):
            with g.tr:
                if prefix:
                    g.td(style=prefix_style, **rowspan)(prefix)
                g.td(style=col_style, **rowspan)(" ")
                g.th(colspan=c - 1, align="left", style=headline_style)(value.get_variant().__name__)
                if col_add:
                    col_add(g, value)
            i = 0
            for f in value.get_variant().fields():
                self.__display_value_internal(g, value[i], col_add, f.name if f.name != "Payload" else None, c - 2,
                                              value.get_description(i))
                i += 1

        elif isinstance(value, ListValue):
            with g.tr:
                if prefix:
                    g.td(style=prefix_style, **rowspan)(prefix)
                g.td(rowspan=l, style=col_style)(" ")
                g.th(colspan=c - 1, align="left", style=headline_style)(value.get_type().__name__)

                if col_add:
                    col_add(g, value)

            for i in range(0, len(value)):
                self.__display_value_internal(g, value[i], col_add, None, c - 1)


        else:
            with g.tr:
                if prefix:
                    if data_span != 1:
                        g.td(colspan=data_span, style=prefix_style)(prefix)
                        data_span = 1
                    else:
                        g.td(style=prefix_style)(prefix)
                else:
                    data_span += 1

                txt = str(value)

                if description is not None:
                    txt = "%s (%s)" % (txt, description)

                if data_span != 1:
                    g.td(colspan=data_span)(txt)
                else:
                    g.td(txt)

                if col_add:
                    col_add(g, value)




#!/usr/bin/env python3

#from ttproto.ts_coap.proto_specific import CoAPTestcase
from ttproto.ts_coap.proto_templates import *
from ttproto.ts_coap.testcases.td_coap_core_19 import TD_COAP_CORE_19


class TD_COAP_CORE_25 (TD_COAP_CORE_19):
    """Identifier:
TD_COAP_CORE_25
Objective:
Perform POST transaction with  responses containing several Location- Query  option
(Reverse proxy)
Configuration:
CoAP_CFG_03
References:
[1] clause 5.8.1,5.10.8,5.9.1.1, 8.2.2,8.2.1,10.2.2,11.2

Pre-test
conditions:
•   Proxy is configured as a reverse-proxy for the server
•   Proxy’s cache is cleared
•   Server accepts creation of new resource on uri  /location-query, the location of
the created resource contains two query parameters ?first=1&second=2

Test Sequence:
Step
Type
Description

1
Stimulus
Client is requested to send a confirmable POST request to
proxy

2
Check
Proxy receives the request from client  & forwards it to
server’s resource


3
Check
Forwarded request must contain:
•   Type = 0 (CON)
•   Code = 2 (POST)
•   An arbitrary payload
•   Content-format option

4
Check
Server sends response to proxy containing:
•	Code = 65 (2.01 created)
•	Two options whose type is Location-Query
	The first option contains first=1
	The second option contains second=2

5
Check
Proxy forwards the response to client

6
Check
Client displays  the message

7
Verify
Client interface returns the response:
•	2.01 created
•	Location: coap://proxy/?first=1&second=2
	"""
    reverse_proxy = True




#!/usr/bin/env python3

from ttproto.ts_coap.proto_specific import CoAPTestcase
from ttproto.ts_coap.proto_templates import *
from ttproto.ts_coap.testcases.td_coap_core_18 import TD_COAP_CORE_18


class TD_COAP_CORE_24(TD_COAP_CORE_18):
    """Identifier:
TD_COAP_CORE_24
Objective:
Perform POST transaction with responses containing several Location-Path options
(Reverse Proxy in CON mode)
Configuration:
CoAP_CFG_03
References:
[1] clause 5.8.1,5.10.8,5.9.1.1, 8.2.2,8.2.1,10.2.2,11.2

Pre-test
conditions:
•	Proxy is configured as a reverse-proxy for the server
•	Proxy’s cache is cleared
•	Server accepts creation of new resource on /create2 and the created resource is
located at  /location1/location2/location3 (resource does not exist yet)


Test Sequence:
Step
Type
Description

1
Stimulus
Client is requested to send a confirmable POST request to
proxy

2
Check
The POST sent by the client contains:
•	Type = 0 (CON)
•	Code = 2 (POST)
•	An arbitrary payload
•	Content-format option


3
Check
The Proxy forwards the POST request to server’s resource
and that it contains:
•	Type = 0 (CON)
•	Code = 2 (POST)
•	An arbitrary payload
•	Content-format option

4
Check
Server sends a response to the proxy containing:
•	Code = 65 (2.01 created)
•	Option type = Location-Path  (one for each segment)
•	Option values must contain “location1”, “location2” &
“location3” without contain any ‘/’


5
Check/
Observe that the Proxy forwards the response (in step 4) to
client and check that the forwarded response contains:
•	Code = 65 (2.01 created)
•	Option type = Location-Path  (one for each segment)
•	Option values must contain “location1”, “location2” &
“location3” without contain any ‘/’

6
Verify
Client displays the response

7
Verify
Client interface returns the response
•	2.01 created
•	Location: coap://proxy/location1/location2/location3
	"""
    request_uri = "/create2"
    reverse_proxy = True



#!/usr/bin/env python3
#
#   (c) 2012  Universite de Rennes 1
#
# Contact address: <t3devkit@irisa.fr>
#
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import os
import sys

from unittest import main, TestLoader, TestResult
from importlib import import_module

TEST_DIR = 'tests'


if __name__ == '__main__':

    sys.stderr.write('################# BEGINNING OF TESTS #################')

    # NOTE: In all the following, write the output on stderr because stdout is
    #       redirected for webservers. By the way, it seems that even the
    #       sys.stderr.write into the tested modules break the redirected pipe.

    # # FIRST WAY

    # # Load all the test cases from the discover method
    # loader = TestLoader()
    # ttproto_test_suite = loader.discover(TEST_DIR)

    # # Run all the test cases providing the result storage
    # result = TestResult()
    # ttproto_test_suite.run(result)

    # # Display results
    # sys.stderr.write("\n\n")
    # sys.stderr.write('################## ERRORS ##################')
    # for err in result.errors:
    #     sys.stderr.write(err[0].__name__ + ': ' + err[1])

    # sys.stderr.write("\n\n")
    # sys.stderr.write('################## FAILURES ##################')
    # for fail in result.failures:
    #     sys.stderr.write(fail[0].__name__ + ': ' + fail[1])

    # sys.stderr.write("\n\n")
    # sys.stderr.write('################## SKIPPED ##################')
    # for skip in result.skipped:
    #     sys.stderr.write(skip[0].__name__ + ': ' + skip[1])

    # SECOND WAY
    # Seek recursively every module into tests directory
    for root, subdirs, files in os.walk(TEST_DIR):

        # Execute every test file found
        for filename in files:
            if all((
                filename.startswith('test'),
                filename.endswith('.py'),
                filename != 'tests.py'  # We don't check old test files
            )):
                test_name = os.path.splitext(filename)[0].upper()
                mod_name = root.replace('/', '.') + '.' + test_name.lower()

                sys.stderr.write("\n\n")
                sys.stderr.write('############ ' + test_name + ' ############')
                sys.stderr.write("\n")

                main(module=mod_name, exit=False)

                sys.stderr.write('###########################################')



#!/usr/bin/env python3
#
#   (c) 2012  Universite de Rennes 1
#
# Contact address: <t3devkit@irisa.fr>
#
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import threading, traceback
from contextlib import contextmanager

from ttproto.core.typecheck	import *
from ttproto.core.data		import *
from ttproto.core		import exceptions, logger, named, snapshot, primitive

__all__ = [
	'Port',
	'BaseMessagePort',
	'RawMessagePort',
	'MessagePort',
	'EventMessageSent',
	'EventMessageReceived',
	'EventMessageMismatch',
]


class Port (named.NamedObject, logger.LoggedObject):
	"""Abstract base class representing a communication port

	Communication ports are used to exchange messages between different
	entities/threads and with the System Under Test.
	"""

	def __init__ (self):
		named.NamedObject.__init__ (self)
		logger.LoggedObject.__init__ (self)

	def __str__ (self):
		return "port:%s" % self.__name__

class BaseMessagePort (Port):
	"""Base class for message-oriented ports


	The BaseMessagePort class handles the forwarding of messages between
	two derived implementations. This process is fully asynchronous. The
	implementation may implement synchronous call on the user side.


            PortImpl_1       BaseMessagePort      PortImpl_2
                 |                  |                  |
         send(…) |                  |                  |
         ------->|  _forward(msg1)  |                  |
                 |----------------->|  enqueue(msg1)   |
         recv(…) |                  |----------------->|
         ------->|                  |                  |
                :|                  |                  |
                :|                  |  _forward(msg2)  |
                :|  enqueue(msg2)   |<-----------------|
                :|<-----------------|                  |
         <------'|                  |                  |
                 |                  |                  |
                 |                  |                  |

	Before exchanging messages, a port shall be connected to another
	MessagePort (using the connect() method).


	NOTE: The current implementation does not allow multiple concurrent
	connections on the same port for the moment.

	"""

	def __init__ (self, endpoint: optional (this_class) = None):
		"""Initialise the port

		If the 'endpoint' parameter is present, this port will be
		immediately connected to this endpoint.
		"""
		Port.__init__ (self)

		self.__endpoint = None
		self.__lock = threading.Lock()
		if endpoint:
			self.connect (endpoint)


	def connect (self, other: this_class):
		"""Connect this port to another port

		The connection is bidirectionnal, and can be broken by any of
		the two ports upon a call to .disconnect().
		"""
		assert isinstance (other, BaseMessagePort)

		if other is self:
			with self.__lock:
				if self.__endpoint:
					raise exceptions.Error("port is already connected")
				self.__endpoint = self

		else:
			la, lb = self.__lock, other.__lock
			if id(la) > id(lb):
				la, lb = lb, la
			with la:
				with lb:
					for p in self, other:
						if p.__endpoint:
							raise exceptions.Error("port is already connected")

					self.__endpoint  = other
					other.__endpoint = self

	def disconnect (self):
		"""Disconnect this port"""
		with self.__lock:
			other = self.__endpoint
			if other is self or other is None:
				self.__endpoint = None
				return

			# we must lock in the right order
			if id(other.__lock) > id(self.__lock):
				with other.__lock:
					self.__endpoint = None
					other.__endpoint = None

		with other.__lock:
			with self.__lock:
				if self.__endpoint == other: # otherwire it means that someone else disconnected it
					self.__endpoint = None
					other.__endpoint = None

	@contextmanager
	def endpoint (self) -> optional (this_class):
		"""Return a context to access the endpoint of this port.

		This contexts the port yields the identity of the endpoint
		connected to this port (or None if not connected).

		When in the context, the port is locked. All other operations
		on the port are blocked until the context is left. Thus the
		endpoint is guaranteed not to change until the context is left.

		Example:
			with my_port.endpoint() as ep:
				print (my_port, "is connected to", ep)

		"""
		with self.__lock:
			yield self.__endpoint


	@typecheck
	def _forward (self, msg: Message) -> bool:
		"""Forward a message to the endpoint of this port

		return False if the port is not connected (else return True)
		"""

		with self.endpoint() as ep:
			if ep is None:
				return False

#			print ("%s->%s: %s" % (self.__name__, ep.__name__, msg))
			ep.enqueue (msg)
			return True

#FIXME: the implementation is not symetrical
class RawMessagePort (BaseMessagePort):

	def __init__ (self, decode_type: is_type = bytes, endpoint = None):
		BaseMessagePort.__init__ (self, endpoint)

		self.__decode_type = get_type (decode_type)


	@typecheck
	def _forward (self, bin_msg: bytes):

		# decode the message
		try:
			msg = Message (bin_msg, self.__decode_type)
		except Exception as e:
			# TODO: log this
			print("Decoding error ->", e)
#			traceback.print_exc()
			msg = Message (bin_msg, primitive.BytesValue)

		return super()._forward (msg)


class MessagePort (snapshot.EventSource, BaseMessagePort):
	"""Message port that can interact with the test

	MessagePort is a message-oriented port that provides a send() function
	to emit a message to the remote port, and a receive(...) event that can
	matched in snapshot blocks.

		Testcase   MessagePort      BaseMessagePort
		    |             |                |
		    |   send(…)   |                |
		    |------------>| _forward(msg1) |
		    |             |--------------->|
		    | receive(…)  |                |
		    |------------>|                |
		    |            :|                |
		    |            :|                |
		    |            :|  enqueue(msg2) |
		    |            :|<---------------|
		    |<-----------'|                |
		    |             |                |
		    |             |                |

	Example:
		link1 = MessagePort()
		link1.connect (some_other_port)

		# synchronous operation (will be blocking)
		link1.send (ping_request_msg)
		link1.receive (ping_request_msg)

		# asynchronous (non-blocking)
		link1.send (ping_request_msg)
		t = Timer()
		t.start (10)
		with alt:
			@link1.receive (ping_request_msg)
			def _():
				set_verdict (pass)

			@link1.receive()
			def _():
				set_verdict (fail)

			@t.timeout()
			def _():
				set_verdict (fail)
	"""

	class ReceiveMatch (snapshot.EventMatch):
		@typecheck
		def __init__ (self, source: snapshot.Event, message: Message, data: optional(is_data)):
			super().__init__(source)
			self.__message = message
			self.__data = store_data (data, none_is_allowed = True)

		def get_message (self):
			return self.__message

		def get_binary (self):
			return self.__message.get_binary()

		def get_value (self):
			return self.__message.get_value()

		def get_data (self):
			return self.__data


	def __init__ (self, endpoint = None):
		snapshot.EventSource.__init__ (self)
		BaseMessagePort.__init__ (self, endpoint)

		self.receive = snapshot.Event (self, "receive")

		self.__queue_in = []
		self.__failed_matches = []

	@typecheck
	def enqueue (self, msg: Message):
		with self.lock:
#			print "%s: enqueue message -> %s" % (self.__name__, msg)
			self.__queue_in.append (msg)

	@typecheck
	def send (self, value: is_value):
		msg = Message (value)
		if not self._forward (msg):
			raise exceptions.PortNotConnected()

		self.log (EventMessageSent, msg, value)

	def evaluate_snapshot (self):
		self.__snapshot_msg = bool (self.__queue_in)

	@typecheck
	def match_receive (self, data: optional (is_data) = None):
		if self.__snapshot_msg:
			assert self.__queue_in

			# TODO: match the template
			msg = self.__queue_in[0]

			mismatch_list = []
			if data is not None and not as_data(data).match(msg.get_value(), mismatch_list):
				if data not in self.__failed_matches:
#					print "%s: mismatch  ->  got: %s, expected: %s" % (self.__name__, msg, data)
					self.__failed_matches.append (data)
					self.log (EventMessageMismatch, msg, data, tuple (mismatch_list))
				return None

#			print("%s: receive message -> %s, expected: %s" % (self.__name__, msg, data))
#			display (msg.get_value())
#			print ("%s:recv:" % self.__name__, msg)

			# TODO: lock the object ??? (is it safe to touch the list now)
			self.__queue_in.pop(0)
			self.__failed_matches[:] = ()

			self.log (EventMessageReceived, msg, data)

			return self.ReceiveMatch (self.receive, msg, data)

		return None

class EventMessageSent (metaclass = logger.LogEventClass):
	fields = (("port", BaseMessagePort),
		  ("message", Message),
		  ("pattern", Data),
		 )
	def summary (self):
		return "%s--> %s" % (self[0].__name__, self[1].summary())

class EventMessageReceived (metaclass = logger.LogEventClass):
	fields = (("port", BaseMessagePort),
		  ("message", Message),
		  ("pattern", Data, True),
		 )
	def summary (self):
		return "%s<-- %s" % (self[0].__name__, self[1].summary())

class EventMessageMismatch (metaclass = logger.LogEventClass):
	fields = (("port", BaseMessagePort),
		  ("message", Message),
		  ("pattern", Data),
		  ("mismatches", tuple)
		 )
	def summary (self):
		return "%s<-- mismatch" % self[0].__name__






import doctest
import unittest
import ttproto.tat_coap.common

# python3 -m unittest tests/test_doctests_common.py -vvv

suite = unittest.TestSuite()
suite.addTest(doctest.DocTestSuite(ttproto.tat_coap.common))
runner = unittest.TextTestRunner(verbosity=2)
runner.run(suite)



#!/usr/bin/env python3
#
#   (c) 2012    Universite de Rennes 1
#
# Contact address: <t3devkit@irisa.fr>
#
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.    You can  use,
# modify and/or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors    have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
import itertools

from ttproto.core.data import Data, Template, get_type, store_data, DifferenceList, Mismatch
from ttproto.core.templates	import Length, Regex, All, Range, Any

from .proto_specific import *

import ttproto.core.lib.inet.all

class Not (Template):
    def __init__ (self, data):
        super().__init__ (get_type (data))
        self.__data = store_data (data)

    def _repr (self):
        return "%s(%s)" % (
            type (self).__name__,
            repr (self.__data),
        )

    def _template_match (self, data):
        return not data in self.__data


class CoAPOptMismatch (Mismatch):
    def __init__ (self, value, pattern, opt_type):
        super().__init__ (value, pattern)
        self.opt_type = opt_type

    def describe_value (self, describe_func = None):
        return ", ".join (str(o) for o in self.value if type (o) is self.opt_type)

    def describe_expected (self, describe_func = None):
        return ", ".join (str(o) for o in self.pattern.opts (self.opt_type))

class Opt (Template):
    """A template to match a list of options in the right order

    Options are matched by type (eg: CoAPOptionToken, CoAPOptionContentFormat, ...)

    All types that are not present in the template will be ignored at
    matching time.

    If a type occurs at least once in the template, then all options of
    this type in the value must be present and match the corresponding
    options in the template, and in the same order.
    """

    def __init__ (self, *opts, superset=False):
        super().__init__ (CoAPOptionList)

        self.__opts = {}
        self.__superset = superset
        for o in opts:

            t = get_type (o)
            assert issubclass (t, CoAPOption)

            try:
                self.__opts[t].append (store_data (o))
            except KeyError:
                self.__opts[t]=[o]

    def opts (self, type):
        lst = self.__opts.get (type)
        return iter (lst if lst else ())

    def _repr (self):
        return "%s(%s%s)" % (
            type (self).__name__,
            ", ".join ((", ".join (repr(v) for v in l)) for l in self.__opts.values()),
            (", superset=True" if self.__superset else "")
        )

    def _match (self, value, mismatch_list):
        end = []

        result = True
        for typ, lst in self.__opts.items():
            ok = True
            for val, tp in itertools.zip_longest (
                    # select all values matching this type
                    filter ((lambda x: isinstance (x, typ)), value),
                    lst, fillvalue = end):
                if val is end:
                    # more templates than values
                    ok = False
                    break
                if tp is end:
                    # more values than templates
                    if not self.__superset:
                        # superset not allowed
                        ok = False
                    break

                if val not in tp:
                    ok = False
            if not ok:
                result = False
                if mismatch_list is not None:
                    mismatch_list.append (CoAPOptMismatch (value, self, typ))
        return result

class CoAPNoOptMismatch (Mismatch):
    def __init__ (self, value, pattern, opt_v, opt_p):
        super().__init__ (value, pattern)
        self.opt_value = opt_v
        self.opt_pattern = opt_p

    def describe_value (self, describe_func = None):
        return str (self.opt_value)

    def describe_expected (self, describe_func = None):
        return "%s  not present" % self.opt_pattern

class NoOpt (Template):
    """A template to check that some options are not present"""

    def __init__ (self, *opts):
        super().__init__ (CoAPOptionList)

        self.__opts = [store_data (o) for o in opts]

    def _repr (self):
        return "%s(%s)" % (
            type (self).__name__,
            ", ".join (repr (o) for o in self.__opts),
        )

    def _match (self, value, mismatch_list):
        for t in self.__opts:
            for v in value:
                if v in t:
                    if mismatch_list is not None:
                        mismatch_list.append (CoAPNoOptMismatch (value, self, v, t))
                    return False
        return True





from ..common import *


class TD_COAP_CORE_08 (CoAPTestCase):
    """
---
TD_COAP_CORE_08:
    cfg: CoAP_CFG_BASIC
    obj: Perform POST transaction (NON mode)
    pre: Server accepts POST request on /test
    ref: '[COAP] 5.8.2, 5.2.3'
    seq:
    -   s:
        - 'Client is requested to send a POST request with:'
        -   - Type = 1 (NON)
            - Code = 2(POST)
            - An arbitrary payload
            - Content-format option
    -   c:
        - 'The request sent by the client contains:'
        -   - Type=1 and Code=2
            - "Client-generated Message ID (\u2794 CMID)"
            - "Client-generated Token (\u2794 CTOK)"
            - Uri-Path option "test"
    -   v: Server displays the received information
    -   c:
        - 'Server sends response containing:'
        -   - Type = 1 (NON)
            - Code = 2.01 (Created) or 2.04 (Changed)
            - "Server-generated Message ID (\u2794 SMID)"
            - Token = CTOK
            - Zero or more Location-path options
            - Content-format option if payload non-empty
            - Empty or non-empty Payload
    -   v: Client displays the received response
    """

    @classmethod
    @typecheck
    def get_stimulis(cls) -> list_of(Value):
        """
        Get the stimulis of this test case. This has to be be implemented into
        each test cases class.

        :return: The stimulis of this TC
        :rtype: [Value]
        """
        # TODO client message w/ no content format and payload -> fail?
        return [
            CoAP(
                type='non',
                code='post',
                pl=Not(b''),
                opt=Opt(CoAPOptionContentFormat())
            )
        ]

    def run (self):
        self.match ("client", CoAP (type="non", code="post",
                        opt = self.uri ("/test")))
        self.match ("client", CoAP (
                        pl  = Not (b''),
                        opt = Opt (CoAPOptionContentFormat()),
                ), "fail")
        CTOK = self.coap["tok"]

        self.next()

        self.match ("server", CoAP (
                        type = "non",
                        code = Any (65, 68),
                        tok = CTOK,
                ))
        if self.match ("server", CoAP(pl = Not(b"")),None):
            self.match ("server", CoAP (
                        opt = Opt (CoAPOptionContentFormat()),
                ), "fail")



#!/usr/bin/env python3
#
#   (c) 2012  Universite de Rennes 1
#
# Contact address: <t3devkit@irisa.fr>
#
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from .common import *
from . import implemR
################################################################################
# Dummy init
################################################################################

@testcase(SixLoWPAN_ND)
def init( link1 ):
	ereq = pack (
		IPv6 (src = "::", dst = "::" ),
		ICMPv6EchoRequest(),
		"init()"
	)
	link1.send(ereq)

################################################################################
# Test_LPND_1.3.1
################################################################################
@testcase(SixLoWPAN_ND)
def Test_LPND_1_3_1(NUT , TIMER1 , HOST1 , LBR1 , PRFX1 , link1, action):

	t = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	rs = pack (
		IPv6( src = HOST1.lladdr , dst = IPV6_ALL_ROUTERS ) /
		ICMPv6RSol( opt = [
			ICMPv6SLL( hw=NUT.hwaddr )
			]
		)
	)

	step (1, "We perform the common test setup 1.4" )
	CommonTestSetup_1_4( LBR1 , HOST1 , NUT, PRFX1 , TIMER1, link1)

	step (2, "We send a Router Solicitation" )	
	link1.send(rs)
	t.start(TIMER1)

	step (3, "We expect a Router Advertisement" )	
	link1.receive (
		IPv6( dst = HOST1.lladdr ) /
		ICMPv6RAdv( opt = superop ( [
			ICMPv6SLL( hw=NUT.hwaddr ) ,
			SixPI( l=0 )
			] )
		)
	)
	set_verdict("pass")

################################################################################
# Test_LPND_1.3.2a
################################################################################
@testcase(SixLoWPAN_ND)
def Test_LPND_1_3_2a(NUT , TIMER1 , HOST1 , LBR1 , PRFX1 , link1, action):

	t = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	rs = pack (
		IPv6( src = HOST1.lladdr , dst = IPV6_ALL_ROUTERS ) /
		ICMPv6RSol( opt = [
			ICMPv6SLL( hw=HOST1.hwaddr )
			]
		)
	)

	step (1, "We perform the common test setup 1.4" )
	CommonTestSetup_1_4( LBR1 , HOST1 , NUT, PRFX1 , TIMER1, link1)

	step (2, "We send a Router Solicitation" )	
	link1.send(rs)
	t.start(TIMER1)

	step (3, "We expect a Router Advertisement" )	
	link1.receive (
		IPv6( src = NUT.lladdr , dst = HOST1.lladdr ) /
		ICMPv6RAdv( )
	)
	set_verdict("pass")

################################################################################
# Test_LPND_1.3.2b
################################################################################
@testcase(SixLoWPAN_ND)
def Test_LPND_1_3_2b(NUT , TIMER1 , HOST1 , LBR1 , PRFX1 , link1, action):

	t = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	rs = pack (
		IPv6( src = HOST1.lladdr , dst = IPV6_ALL_ROUTERS ) /
		ICMPv6RSol( opt = [
			ICMPv6SLL( hw=HOST1.hwaddr )
			]
		)
	)

	step (1, "We perform the common test setup 1.4" )
	CommonTestSetup_1_4( LBR1 , HOST1 , NUT, PRFX1 , TIMER1, link1)

	step (2, "We send a Router Solicitation" )	
	link1.send(rs)
	t.start(TIMER1)

	step (3, "We expect a Router Advertisement" )	
	link1.receive (
		IPv6( src = NUT.lladdr , dst = HOST1.lladdr ) /
		ICMPv6RAdv(  opt = superop ( [
			ICMPv6SLL( hw=NUT.hwaddr )
			] )
		)
	)
	set_verdict("pass")



################################################################################
# Test_LPND_1.3.2c
################################################################################
@testcase(SixLoWPAN_ND)
def Test_LPND_1_3_2c(NUT , TIMER1 , HOST1 , LBR1 , PRFX1 , link1, action):

	t = Timer()
	t2 = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	rs = pack (
		IPv6( src = HOST1.lladdr , dst = IPV6_ALL_ROUTERS ) /
		ICMPv6RSol( opt = [
			ICMPv6SLL( hw=HOST1.hwaddr )
			]
		)
	)

	ereq = pack (
		IPv6 (src = HOST1.addr, dst = NUT.addr ),
		ICMPv6EchoRequest(),
		"6lowpan-nd-1.3.2c"
	)

	step (1, "We perform the common test setup 1.4" )
	CommonTestSetup_1_4( LBR1 , HOST1 , NUT, PRFX1 , TIMER1, link1)

	step (2, "NODE1 sends an ICMPv6 Router Solicitation" )
	link1.send(rs)

	step (3, "We send an ICMPv6 Echo Request and don't expect any answer" )
	link1.send(ereq)
	t2.start(TIMER1)
	with alt:

		@link1.receive ( IPv6( src=NUT.addr , dst = HOST1.addr )/ICMPv6EchoReply() )
		def _( ):
			set_verdict ("fail")

		@t2.timeout ()
		def _():
			set_verdict ("pass")


################################################################################
# Test_LPND_1.3.3a
################################################################################
@testcase(SixLoWPAN_ND)
def Test_LPND_1_3_3a(NUT , TIMER1 , HOST1 , LBR1 , PRFX1 , link1, action):

	t = Timer()

	unexpected_packets.activate( link1 )

	ns = pack (
		IPv6( src = HOST1.addr , dst = NUT.lladdr ) ,
		ICMPv6NSol( opt = [
			ICMPv6SLL( hw=HOST1.hwaddr ) ,
			SixARO( Length = 9 , eui = HOST1.eui64 )
			]
		)
	)

	step (1, "We perform the common test setup 1.4" )
	CommonTestSetup_1_4( LBR1 , HOST1 , NUT, PRFX1 , TIMER1, link1)

	step (2, "We send an Neighboor Solicitation and don't expect any Answer" )
	link1.send(ns)
	t.start(TIMER1)

	with alt:
		@link1.receive (
			IPv6( src = NUT.lladdr , dst = HOST1.lladdr ) /
			ICMPv6NAdv( )
		)
		def _():
			set_verdict ("fail")
		@t.timeout ()
		def _():
			set_verdict ("pass")

################################################################################
# Test_LPND_1.3.3b
################################################################################
@testcase(SixLoWPAN_ND)
def Test_LPND_1_3_3b(NUT , TIMER1 , HOST1 , LBR1 , PRFX1 , link1, action):

	t = Timer()

	unexpected_packets.activate( link1 )

	ns = pack (
		IPv6( src = HOST1.addr , dst = NUT.lladdr ) ,
		ICMPv6NSol( opt = [
			ICMPv6SLL( hw=HOST1.hwaddr ) ,
			SixARO( st = 80 , eui = HOST1.eui64 )
			]
		)
	)

	step (1, "We perform the common test setup 1.4" )
	CommonTestSetup_1_4( LBR1 , HOST1 , NUT, PRFX1 , TIMER1, link1)

	step (2, "We send an Neighboor Solicitation and don't expect any Answer" )
	link1.send(ns)
	t.start(TIMER1)

	with alt:
		@link1.receive (
			IPv6( src = NUT.lladdr , dst = HOST1.lladdr ) /
			ICMPv6NAdv( )
		)
		def _():
			set_verdict ("fail")
		@t.timeout ()
		def _():
			set_verdict ("pass")

################################################################################
# Test_LPND_1.3.3c
################################################################################
@testcase(SixLoWPAN_ND)
def Test_LPND_1_3_3c(NUT , TIMER1 , HOST1 , LBR1 , PRFX1 , link1, action):

	t = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	ns = pack (
		IPv6( src = HOST1.addr , dst = NUT.lladdr ) ,
		ICMPv6NSol( opt = [
			ICMPv6SLL( hw=HOST1.hwaddr ) ,
			SixARO( st = 0 , eui = HOST1.eui64 )
			]
		)
	)

	step (1, "We perform the common test setup 1.4" )
	CommonTestSetup_1_4( LBR1 , HOST1 , NUT, PRFX1 , TIMER1, link1)

	step (2, "We send an Neighboor Solicitation and a expect Neighboor advertisement without any ARO" )
	link1.send(ns)
	t.start(TIMER1)

	with alt:
		@link1.receive (
			IPv6( src = NUT.lladdr , dst = HOST1.lladdr ) /
			ICMPv6NAdv( opt = superop ( [
				SixARO(  )
				] )
			)
		)
		def _():
			set_verdict ("fail")

		@link1.receive (
			IPv6( src = NUT.lladdr , dst = HOST1.lladdr ) /
			ICMPv6NAdv( opt = [] )
		)
		def _():
			set_verdict ("pass")

################################################################################
# Test_LPND_1.3.3d
################################################################################
@testcase(SixLoWPAN_ND)
def Test_LPND_1_3_3d(NUT , TIMER1 , HOST1 , LBR1 , PRFX1 , link1, action):

	t = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	ns = pack (
		IPv6( src = HOST1.addr , dst = NUT.lladdr ) ,
		ICMPv6NSol( opt = [
			SixARO( eui = HOST1.eui64 )
			]
		)
	)

	step (1, "We perform the common test setup 1.4" )
	CommonTestSetup_1_4( LBR1 , HOST1 , NUT, PRFX1 , TIMER1, link1)

	step (2, "We send an Neighboor Solicitation and a expect Neighboor advertisement without any ARO" )
	link1.send(ns)
	t.start(TIMER1)

	with alt:
		@link1.receive (
			IPv6( src = NUT.lladdr , dst = HOST1.lladdr ) /
			ICMPv6NAdv( opt = superop ( [
				SixARO( )
				] )
			)
		)
		def _():
			set_verdict ("fail")

		@link1.receive (
			IPv6( src = NUT.lladdr , dst = HOST1.lladdr ) /
			ICMPv6NAdv( opt = [] )
		)
		def _():
			set_verdict ("pass")

################################################################################
# Test_LPND_1.3.4a
################################################################################
@testcase(SixLoWPAN_ND)
def Test_LPND_1_3_4a(NUT , TIMER1 , HOST1 , LBR1 , PRFX1 , link1, action):

	t = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	ns = pack (
		IPv6( src = HOST1.addr , dst=NUT.lladdr ) ,
		ICMPv6NSol( opt = [
			SixARO( eui = HOST1.eui64 ),
			ICMPv6SLL( hw=HOST1.hwaddr ) 
			]
		)
	)

	step (1, "We perform the common test setup 1.4" )
	CommonTestSetup_1_4( LBR1 , HOST1 , NUT, PRFX1 , TIMER1, link1)

	step (2, "We send an Neighboor Solicitation and a expect Neighboor advertisement without any ARO" )
	link1.send(ns)
	t.start(TIMER1)

	link1.receive (
			IPv6( src = NUT.lladdr ) /
			ICMPv6NAdv( opt = superop ( [
				SixARO(  st = 0  , eui = HOST1.eui64 )
				] )
			)
		)
	set_verdict ("pass")

################################################################################
# Test_LPND_1.3.4b
################################################################################
@testcase(SixLoWPAN_ND)
def Test_LPND_1_3_4b(NUT , TIMER1 , HOST1 , LBR1 , PRFX1 , link1, action):

	t = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	ns = pack (
		IPv6( src = PRFX1+"42:1" , dst = NUT.lladdr ) ,
		ICMPv6NSol( opt = [
			SixARO( eui = "1234567890000000" ) ,
			ICMPv6SLL( hw=HOST1.hwaddr )
			]
		)
	)

	step (1, "We perform the common test setup 1.4" )
	CommonTestSetup_1_4( LBR1 , HOST1 , NUT, PRFX1 , TIMER1, link1)

	step (2, "We send an Neighboor Solicitation and a expect Neighboor advertisement with an ARO with a st = 0 " )
	link1.send(ns)
	t.start(TIMER1)

	link1.receive (
			IPv6( src = NUT.lladdr ) /
			ICMPv6NAdv( opt = superop ( [
				SixARO(  st = 0  , eui = "1234567890000000")
				] )
			)
		)
	set_verdict ("pass")

	step (3, "We send an Neighboor Solicitation and a expect Neighboor advertisement with an ARO with a st = 0 " )
	link1.send(ns)
	t.start(TIMER1)

	link1.receive (
			IPv6( src = NUT.lladdr ) /
			ICMPv6NAdv( opt = superop ( [
				SixARO(  st = 0  , eui = "1234567890000000" )
				] )
			)
		)
	set_verdict ("pass")

################################################################################
# Test_LPND_1.3.4c
################################################################################
@testcase(SixLoWPAN_ND)
def Test_LPND_1_3_4c(NUT , TIMER1 , HOST1 , HOST2, LBR1 , PRFX1 , link1, action):

	t = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	ns = pack (
		IPv6( src = PRFX1+"42:1" , dst = NUT.lladdr ) ,
		ICMPv6NSol( opt = [
			SixARO( eui = "1234567890000000" ) ,
			ICMPv6SLL( hw=HOST1.hwaddr )
			]
		)
	)

	ns2 = pack (
		IPv6( src = PRFX1+"42:1" , dst = NUT.lladdr ) ,
		ICMPv6NSol( opt = [
			SixARO( eui = "1234567890000011" ),
			ICMPv6SLL( hw=HOST2.hwaddr ) 
			]
		)
	)

	step (1, "We perform the common test setup 1.4" )
	CommonTestSetup_1_4( LBR1 , HOST1 , NUT, PRFX1 , TIMER1, link1)

	step (2, "We send an Neighboor Solicitation and a expect Neighboor advertisement with an ARO with a st = 0 " )
	link1.send(ns)
	t.start(TIMER1)

	link1.receive (
			IPv6( src = NUT.lladdr ) /
			ICMPv6NAdv( opt = superop ( [
				SixARO(  st = 0  , eui = "1234567890000000" )
				] )
			)
		)
	set_verdict ("pass")

	step (3, "We send an Neighboor Solicitation and a expect Neighboor advertisement with an ARO with a st = 1 " )
	link1.send(ns2)
	t.start(TIMER1)

	link1.receive (
			IPv6( src = NUT.lladdr ) /
			ICMPv6NAdv( opt = superop ( [
				SixARO(  st = 1  , eui = "1234567890000011" )
				] )
			)
		)
	set_verdict ("pass")


################################################################################
# Test_LPND_1.3.5a
################################################################################
@testcase(SixLoWPAN_ND)
def Test_LPND_1_3_5a(NUT , TIMER1 , HOST1 , HOST2 , LBR1 , PRFX1 , link1, action):

	t = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	ns = pack (
		IPv6( src = PRFX1+"42:1" , dst = NUT.lladdr ) ,
		ICMPv6NSol( opt = [
			SixARO( eui = "1234567890000000" ),
			ICMPv6SLL( hw=HOST1.hwaddr ) 
			]
		)
	)

	ns2 = pack (
		IPv6( src = PRFX1+"42:1" , dst = NUT.lladdr ) ,
		ICMPv6NSol( opt = [
			SixARO( eui = "1234567890000011" ) ,
			ICMPv6SLL( hw=HOST2.hwaddr )
			]
		)
	)

	step (1, "We perform the common test setup 1.4" )
	CommonTestSetup_1_4( LBR1 , HOST1 , NUT, PRFX1 , TIMER1, link1)

	step (2, "We send an Neighboor Solicitation and a expect Neighboor advertisement with an ARO with a st = 0 " )
	link1.send(ns)
	t.start(TIMER1)

	link1.receive (
			IPv6( src = NUT.lladdr ) /
			ICMPv6NAdv( opt = superop ( [
				SixARO(  st = 0  , eui = "1234567890000000" )
				] )
			)
		)
	set_verdict ("pass")

	step (3, "We send an Neighboor Solicitation and a expect Neighboor advertisement with the IPv6 destination field set to " +HOST2.lladdr)
	link1.send(ns2)
	t.start(TIMER1)

	link1.receive (
			IPv6( src = NUT.lladdr , dst = "" ) / #TODO: dst address derived from EUI64
			ICMPv6NAdv( opt = superop ( [
				SixARO()
				] )
			)
		)
	set_verdict ("pass")

################################################################################
# Test_LPND_1.3.5b
################################################################################
@testcase(SixLoWPAN_ND)
def Test_LPND_1_3_5a(NUT , TIMER1 , HOST1 , HOST2 , LBR1 , PRFX1 , link1, action):

	t = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	ns = pack (
		IPv6( src = PRFX1+"42:1" , dst = NUT.lladdr ) ,
		ICMPv6NSol( opt = [
			SixARO( eui = "1234567890000000" ),
			ICMPv6SLL( hw=HOST1.hwaddr ) 
			]
		)
	)

	ns2 = pack (
		IPv6( src = PRFX1+"42:1" , dst = NUT.lladdr ) ,
		ICMPv6NSol( opt = [
			SixARO( eui = "1234567890000011" , lt = 38) ,
			ICMPv6SLL( hw=HOST2.hwaddr )
			]
		)
	)

	step (1, "We perform the common test setup 1.4" )
	CommonTestSetup_1_4( LBR1 , HOST1 , NUT, PRFX1 , TIMER1, link1)

	step (2, "We send an Neighboor Solicitation and a expect Neighboor advertisement with an ARO with a st = 0 " )
	link1.send(ns)
	t.start(TIMER1)

	link1.receive (
			IPv6( src = NUT.lladdr ) /
			ICMPv6NAdv( opt = superop ( [
				SixARO(  st = 0  , eui = "1234567890000000" )
				] )
			)
		)
	set_verdict ("pass")

	step (3, "We send an Neighboor Solicitation and a expect Neighboor advertisement with the lifetime field set to 38" )
	link1.send(ns2)
	t.start(TIMER1)

	link1.receive (
			IPv6( src = NUT.lladdr , dst = "" ),
			ICMPv6NAdv(  opt = superop ( [
				SixARO( st = 1 , eui = "1234567890000011" , lt = 38)
				] )
			)
		)
	set_verdict ("pass")


################################################################################
# Test_LPND_1.3.6a
################################################################################
@testcase(SixLoWPAN_ND)
def Test_LPND_1_3_6a(NUT , TIMER1 , HOST1 , LBR1 , PRFX1 , link1, action):


	t = Timer()
	t2 = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	ns1 = pack (
		IPv6( src = HOST1.addr , dst = NUT.lladdr ) ,
		ICMPv6NSol( opt = [
			ICMPv6SLL( hw=HOST1.hwaddr ) ,
			SixARO(lt = 1 , eui = HOST1.eui64  )
			]
		)
	)
	
	ns2 = pack (
		IPv6( src = HOST1.addr , dst = NUT.lladdr ) ,
		ICMPv6NSol( opt = [
			ICMPv6SLL( hw=HOST1.hwaddr ) ,
			SixARO(lt = 0 , eui = HOST1.eui64  )
			]
		)
	)
	
	ereq1 = pack (
		IPv6 (src = NUT.addr, dst = HOST1.addr ),
		ICMPv6EchoRequest(),
		"6lowpan-nd-1.3.6a-1"
	)

	ereq2 = pack (
		IPv6 (src = NUT.addr, dst = HOST1.addr ),
		ICMPv6EchoRequest(),
		"6lowpan-nd-1.3.6a-2"
	)

	step (1, "We perform the common test setup 1.4" )
	CommonTestSetup_1_4( LBR1 , HOST1 , NUT, PRFX1 , TIMER1, link1)

	step (2, "We send a Neighboor Solicitation with an ARO with a lifetime set to 1 (60 seconds) and we expect a Neighboor Solicitation" )	
	link1.send(ns1)
	t.start(TIMER1)
	link1.receive (
		IPv6( src = NUT.lladdr ) /
		ICMPv6NAdv( opt = superop ( [
			SixARO(  st = 0  , lt = 1 ,  eui = HOST1.eui64  )
			] )
		)
	)
	set_verdict("pass")

	step (3, "We sends an ICMPv6 Echo Request. We expect an ICMPv6 Echo Reply" )
	link1.send(ereq1)
	t.start(TIMER1)
	link1.receive (	IPv6( src=NUT.addr , dst = HOST1.addr ) / ICMPv6ERep() )
	set_verdict("pass")


	step (4, "We send a Neighboor Solicitation with an ARO with a lifetime set to 0 (0 seconds) and  we expect a Neighboor Solicitation" )	
	link1.send(ns2)
	t.start(TIMER1)
	link1.receive (
		IPv6( src = NUT.lladdr ) /
		ICMPv6NAdv( opt = superop ( [
			SixARO(  st = 0  , lt = 0 , eui = HOST1.eui64 ) 
			] )
		)
	)
	set_verdict("pass")

	step (5, "We sends an ICMPv6 Echo Request. We expect an no ICMPv6 Echo Reply ")
	link1.send(ereq2)
	t.stop()
	t2.start(TIMER2)

	with alt:
		@link1.receive( IPv6( src = NUT.addr , dst= HOST1.addr ) / ICMPv6ERep() )
		def _():
			set_verdict("fail")

		@t2.timeout()
		def _():
			set_verdict("pass")


################################################################################
# Test_LPND_1.3.7a
################################################################################
@testcase(SixLoWPAN_ND)
def Test_LPND_1_3_7a(NUT , TIMER1 , HOST1 , LBR1 , PRFX1 , link1, action):

	t = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	ns = pack (
		IPv6( src = HOST1.addr , dst = NUT.lladdr ) ,
		ICMPv6NSol( opt = [
			ICMPv6SLL( hw=HOST1.hwaddr ) ,
			SixARO(lt = 1 ,  eui = HOST1.eui64  )
			]
		)
	)

	step (1, "We perform the common test setup 1.4" )
	CommonTestSetup_1_4( LBR1 , HOST1 , NUT, PRFX1 , TIMER1, link1)

	step (2, "We send a Neighboor Solicitation with an ARO we expect a Neighboor Solicitation" )	
	link1.send(ns)
	t.start(TIMER1)
	link1.receive (
		IPv6( src = NUT.lladdr ) /
		ICMPv6NAdv( opt = superop ( [
			SixARO(  st = 0  , lt = 0 , eui = HOST1.eui64 )
			] )
		)
	)
	set_verdict("pass")

	step (3, "NUT action")
	action ("Order the NUT to emit an ICMPv6 Echo Request to %s" % HOST1.addr)

	t.start(TIMER1)
	step (4, "We expect an ICMPv6 Echo Request with the link layer address set to HOST1" )
	link1.receive (	IPv6( src = NUT.addr , dst=HOST1.addr) / ICMPv6EReq() ,	HOST1.hwaddr )
	set_verdict("pass")


################################################################################
# Test_LPND_1.3.7b
################################################################################
@testcase(SixLoWPAN_ND)
def Test_LPND_1_3_7b(NUT , TIMER1 , HOST1 , LBR1 , PRFX1 , link1, action):

	t = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	step (1, "We perform the common test setup 1.4" )
	CommonTestSetup_1_4( LBR1 , HOST1 , NUT, PRFX1 , TIMER1, link1)

	step (2, "NUT action")
	action ("Order the NUT to emit an ICMPv6 Echo Request to %s" % HOST1.addr)
	
	step (3, "We expect an ICMPv6 Echo Request with the link layer address set to LBR1" ) #TODO: LBR: case ?
	t.start(TIMER1)
	link1.receive (	IPv6( src = NUT.addr , dst=HOST1.addr) / ICMPv6EReq() , LBR1.hwaddr )
	set_verdict("pass")


################################################################################
# Test_LPND_1.3.7c
################################################################################
@testcase(SixLoWPAN_ND)
def Test_LPND_1_3_7c(NUT , TIMER1 , HOST1 , LBR1 , PRFX1 , link1, action):

	t = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	ns = pack (
		IPv6( src = HOST1.addr , dst = NUT.lladdr ) ,
		ICMPv6NSol( opt = [
			ICMPv6SLL( hw=HOST1.hwaddr ) ,
			SixARO(lt = 1 ,  eui = HOST1.eui64  )
			]
		)
	)

	step (1, "We perform the common test setup 1.4" )
	CommonTestSetup_1_4( LBR1 , HOST1 , NUT, PRFX1 , TIMER1, link1)

	step (2, "We send a Neighboor Solicitation with an ARO we expect a Neighboor Advertisement with an ARO" )	
	link1.send(ns)
	t.start(TIMER1)
	link1.receive (
		IPv6( src = NUT.lladdr ) /
		ICMPv6NAdv( opt = superop ( [
			SixARO(  st = 0  , lt = 1 , eui = HOST1.eui64 )
			] )
		)
	)
	set_verdict("pass")

	step (3, "We wait for 60 seconds" )
	t.start (60)
	t.timeout()

	step (4, "NUT action")
	action ("Order the NUT to emit an ICMPv6 Echo Request to %s" % HOST1.addr)

	t.start(TIMER1)
	step (5, "We expect an ICMPv6 Echo Request with the link layer address set to LBR1" )  #TODO: LBR: case ?
	link1.receive (	IPv6( src = NUT.addr , dst=HOST1.addr) / ICMPv6EReq() , LBR1.hwaddr )
	set_verdict("pass")

################################################################################
# Test_LPND_1.3.8a
################################################################################
@testcase(SixLoWPAN_ND)
def Test_LPND_1_3_8a(NUT , TIMER1 , HOST1 , LBR1 , PRFX1 , link1, action):

	t = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	ereq = pack (
		IPv6 (src = NUT.lladdr, dst = HOST1.lladdr ),
		ICMPv6EchoRequest(),
		"6lowpan-nd-1.3.8a"
	)

	step (1, "We perform the common test setup 1.4" )
	CommonTestSetup_1_4( LBR1 , HOST1 , NUT, PRFX1 , TIMER1, link1)

	step (2, "HOST1 sends an ICMPv6 Echo Request to the solicited-node multicast address of the NUT" )
	link1.send(ereq)

	step (3, "We expect an ICMPv6 Echo Reply" )
	t.start(TIMER1)
	link1.receive(
		IPv6 ( src = NUT.lladdr , dst = HOST1.lladdr )  /
		ICMPv6ERep()
		)
	set_verdict("pass")
	

################################################################################
# Test_LPND_1.3.8b
################################################################################
@testcase(SixLoWPAN_ND)
def Test_LPND_1_3_8b(NUT , TIMER1 , HOST1 , LBR1 , PRFX1 , link1, action):

	t = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	ns = pack (
		IPv6( src = HOST1.addr , dst = NUT.lladdr ) ,
		ICMPv6NSol( opt =  [
			ICMPv6SLL( hw=HOST1.hwaddr ) ,
			]
		)
	)

	step (1, "We perform the common test setup 1.4" )
	CommonTestSetup_1_4( LBR1 , HOST1 , NUT, PRFX1 , TIMER1, link1)

	step (2, "HOST1 sends an ICMPv6 Neigboor solicitation to the solicited-node multicast address of the NUT" )
	link1.send(ns)

	step (3, "We expect an ICMPv6 Neigboor advertisement" )
	t.start(TIMER1)
	link1.receive(
		IPv6 ( src = NUT.lladdr , dst = HOST1.addr )  /
		ICMPv6NAdv()
		)
	set_verdict("pass")


################################################################################
# Test_LPND_1.3.9
################################################################################
@testcase(SixLoWPAN_ND)
def Test_LPND_1_3_9(NUT , TIMER1 , HOST1 , LBR1 , PRFX1 , link1, action):

	t = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	ns = pack (
		IPv6( src = HOST1.addr , dst = NUT.lladdr ) ,
		ICMPv6NSol( opt = [
			ICMPv6SLL( hw=HOST1.hwaddr ) ,
			SixARO(lt = 1 ,  eui = HOST1.eui64  )
			]
		)
	)

	step (1, "We perform the common test setup 1.4" )
	CommonTestSetup_1_4( LBR1 , HOST1 , NUT, PRFX1 , TIMER1, link1)

	step (2, "We send a neighboor solicitation with an Address Registration Option (ARO)" )
	link1.send(ns)

	step (2, "We expect a neighboor advertisement with a valid Address Registration Option (ARO)" )
	t.start(TIMER1)

	link1.receive (
		IPv6( src = NUT.lladdr , dst = HOST1.addr ) /
		ICMPv6NAdv( opt = superop ( [
			SixARO(
				type = 33 , #FIXME: temp value, will be allocated
				Length = 2 ,
				Status = 0 ,
				Reserved = 0 ,
				eui = HOST1.eui64
				)
			] )
		)
	)
	set_verdict("pass")


################################################################################
# Test_LPND_1.3.10
################################################################################
@testcase(SixLoWPAN_ND)
def Test_LPND_1_3_10(NUT , TIMER1 , HOST1 , LBR1, PRFX1 , link1, action):

	t = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	rs = pack (
		IPv6( src = HOST1.lladdr , dst = NUT.lladdr ) /
		ICMPv6RSol( opt = [
			ICMPv6SLL( hw=HOST1.hwaddr )
			]
		)
	)
	step (1, "We perform the common test setup 1.4b" )
	CommonTestSetup_1_4( LBR1 , HOST1 , NUT, PRFX1 , TIMER1, link1)

	step (2, "We send a router solicitation" )
	link1.send(rs)

	step (2, "We expect a router advertisement with a valid Compress Option" )
	t.start(TIMER1)

	link1.receive (
		IPv6( src = NUT.lladdr , dst = HOST1.addr ) /
		ICMPv6RAdv( opt = superop ( [
			SixCO (
				type = 32 ,  #FIXME: temp value, will be allocated
				len = 3 ,
				clen = 64 ,
				rsv1 = 0,
				c = 1,
				cid = 1,
				rsv2 = 0,
				lt = 1,
				pf = PRFX1
				)
			] )
		)
	)
	set_verdict("pass")

################################################################################
# Test_LPND_1.3.11
################################################################################
@testcase(SixLoWPAN_ND)
def Test_LPND_1_3_11(NUT , TIMER1 , HOST1 , LBR1, PRFX1 , link1, action):

	t = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	rs = pack (
		IPv6( src = HOST1.lladdr , dst = NUT.lladdr ) /
		ICMPv6RSol( opt = [
			ICMPv6SLL( hw=HOST1.hwaddr )
			]
		)
	)
	step (1, "We perform the common test setup 1.4b" )
	CommonTestSetup_1_4( LBR1 , HOST1 , NUT, PRFX1 , TIMER1, link1)

	step (2, "We send a router solicitation" )
	link1.send(rs)

	step (3, "We expect a router advertisement with a valid Authoritative Border Router Option" )
	t.start(TIMER1)

	link1.receive (
		IPv6( src = NUT.lladdr , dst = HOST1.addr ) /
		ICMPv6RAdv( opt = superop ( [
			SixABRO (
				len = 3 ,
				lbr = LBR1.addr
				)
			] )
		)
	)
	set_verdict("pass")
################################################################################
# Main()
################################################################################

if __name__ == "__main__":
	print("Internal use only. Use testslr.py or testslbr.py")
	exit(1)



from ttproto.tat_coap.testcases.td_coap_core_23 import TD_COAP_CORE_23


class TD_COAP_CORE_28 (TD_COAP_CORE_23):
    """
---
TD_COAP_CORE_28:
    cfg: CoAP_CFG_03
    obj: Perform GET transaction with responses containing the ETag option and
         requests containing the If-None-Match option (CON mode)
         (Reverse proxy)
    pre:
        - Proxy is configured as a reverse-proxy for the server
        - Proxy’s cache is cleared
        - Server should offer a /create3 resource, which does not exist and
          which can be created by the client
        - Client & server supports If-None-Match
    ref: '[1] clause 5.8.1, 5.10.7,5.10.10,12.1.12, 8.2.2,8.2.1,10.2.2,11.2'
    seq:
    -   s: Client is requested to send a confirmable PUT request to proxy to
           atomically create resource in server
    -   c: Proxy forwards the request to server
    -   c:
        - 'Forwarded t request must contain:'
        -   - Type = 0 (CON)
            - Code = 3 (PUT)
            - Option Type=If-None-Match
            - An arbitrary payload
    -   c:
        - 'Server sends response containing:'
        -   - Code = 65 (2.01 Created)
    -   c: Proxy forwards the response to client
    -   v: Client displays the response & and server created new resource
    -   s: Client is requested to send s confirmable PUT request to proxy to
           atomically create resource in server
    -   c:
        - 'Sent request must contain:'
        -   - Type = 0 (CON)
            - Code = 3 (PUT)
            - Option Type=If-Non-Match
            - Option value=Received ETag value
    -   c:
        - 'Server sends response containing:'
        -   - 140 (4.12 Precondition Failed)
    -   v: Proxy forwards the response to client
    -   c:
        - 'Response contains:'
        -   - 140 (4.12 Precondition Failed)
    -   v: Client displays the response
    """

    request_uri = "/create3"
    reverse_proxy = True



#!/usr/bin/env python3

from ttproto.ts_coap.proto_specific import CoAPTestcase
from ttproto.ts_coap.proto_templates import *


class TD_COAP_CORE_22(CoAPTestcase):
    """Identifier:
TD_COAP_CORE_22
Objective:
Perform GET transaction with responses containing the ETag option and requests
containing the If-Match option (CON mode)
Configuration:
CoAP_CFG_BASIC
References:
[COAP] 5.8.1, 5.10.7,5.10.9,12.1.12

Pre-test
conditions:
•	Server offer a /validate resource
•	Client & server supports ETag and If-Match option
•	The Client‘s cache must be purged

Test Sequence:
Step
Type
Description
Preamble: client gets the resource


1
Stimulus
Client is requested to send a confirmable GET request to
server’s resource


2
Check
The request sent by the client contains:
•	Type = 0 (CON)
•	Code = 1 (GET)
•	Client-generated Message ID (➔ CMID)
•	Client-generated Token (➔ CTOK)
•	Uri-Path option "validate"
•	No ETag option

3
Check
Server sends response containing:
•	Code = 69 (2.05 content)
•	Message ID = CMID, Token = CTOK
•	Option type = ETag, value = a value chosen by the server (➔ ETAG1)
•	Non-empty Payload

Part A: single update

4
Stimulus
Client is requested to send a confirmable PUT request to
server’s resource so as to perform an atomic update

5
Check
The request sent by the client contains:
•	Type = 0 (CON)
•	Code = 3 (PUT)
•	Another client-generated Message ID ≠ CMID (➔ CMID2)
•	Client-generated Token which may or may not be ≠ CTOK (➔ CTOK2)
•	Content-format option
•	Uri-Path option "validate"
•	Option type = If-Match, value = ETAG1 (ETag value received in step 3)
•	An arbitrary payload (which differs from the payload received in step 3)

6
Check
Server sends response containing:
•	Code = 68 (2.04 Changed)
•	Message ID = CMID2, Token = CTOK2
•	Content-format option if payload non-empty
•	Empty or non-empty Payload

7
Verify
Client displays the response and the server changed its
resource

Part B: concurrent updates


8
Stimulus
Client is requested to send a confirmable GET request to
server’s resource

9
Check
The request sent by the client contains:
•	Type = 0 (CON)
•	Code = 1 (GET)
•	Another client-generated Message ID ≠ CMID and ≠ CMID2 (➔ CMID3)
•	Client-generated Token which may or may not be ≠ CTOK or CTOK2 (➔ CTOK3)
•	Uri-Path option "validate"

10
Check
Server sends response containing:
•	Code = 69 (2.05 content)
•	Message ID = CMID3, Token = CTOK3
•	Option type = ETag, value = a value ≠ ETAG1 chosen by the server (➔ ETAG2)
•	The Payload sent in step 5

11
Verify
Client displays the response


12
Stimulus
Update the content of the server’s resource from a CoAP client

13
Check
The request sent by the client contains:
•	Type = 0 (CON)
•	Code = 3 (PUT)
•	Another client-generated Message ID ≠ CMID, CMID2, CMID3 (➔ CMID4)
•	Client-generated Token which may or may not be ≠ CTOK, CTOK2, CTOK3 (➔ CTOK4)
•	Content-format option
•	Uri-Path option "validate"
•	An arbitrary payload (which differs from the payloads received in steps 3 and 10)

14
Check
Server sends response containing:
•	Code = 2.04 (Changed)
•	Message ID = CMID4, Token = CTOK4
•	Content-format option if payload non-empty
•	Empty or non-empty Payload

15
Verify
Client displays the response and the server changed its resource

16
Stimulus
Client is requested to send a confirmable PUT request to server’s resource so as to perform an atomic update, assuming it is still unchanged from step 10

17
Check
The request sent by the client contains:
•	Type = 0 (CON)
•	Code = 3 (PUT)
•	Another client-generated Message ID ≠ CMID, CMID2, CMID3, CMID4 (➔ CMID5)
•	Client-generated Token which may or may not be ≠ CTOK, CTOK2, CTOK3, CTOK4 (➔ CTOK5)
•	Content-format option
•	Uri-Path option "validate"
•	Option type = If-Match, value = ETAG2 (ETag value received in step 10)
•	An arbitrary payload (which differs from the previous payloads)

18
Check
Server sends response containing:
•	Code = 4.12 (Precondition Failed)
•	Message ID = CMID4, Token = CTOK4
•	Optional Content-format option
•	Empty or non-empty Payload

19
Verify
Client displays the response and the server did not update the content of the resource
"""

    def run(self):
        # Preamble
        # Step 2
        self.match_coap("client", CoAP(type="con", code="get",
                                       opt=All(
                                           Opt(CoAPOptionUriPath("validate")),
                                           NoOpt(CoAPOptionETag()),
                                       )))
        CMID = self.frame.coap["mid"]
        CTOK = self.frame.coap["tok"]

        self.next_skip_ack()

        # Step 3
        if not self.match_coap("server", CoAP(type=Any(CoAPType("con"), "ack"),
                                              code=2.05,
                                              mid=CMID,
                                              tok=CTOK,
                                              opt=Opt(CoAPOptionETag()),
                                              pl=Not(b""))):
            raise self.Stop()

        ETAG1 = self.frame.coap["opt"][CoAPOptionETag]["val"]
        pl_3 = self.frame.coap["pl"]

        self.next_skip_ack(optional=True)

        # Part A
        self.chain()
        # Step 5
        self.match_coap("client", CoAP(type="con", code="put",
                                       opt=Opt(
                                           CoAPOptionContentFormat(),
                                           CoAPOptionUriPath("validate"),
                                           CoAPOptionIfMatch(ETAG1),
                                       ),
                                       pl=All(Not(b""), Not(pl_3))))
        CMID2 = self.frame.coap["mid"]
        CTOK2 = self.frame.coap["tok"]
        if CMID2 is Not(b''):
            if CMID2 == CMID:
                self.setverdict("fail", "Message ID should be different")
        if CTOK2 is Not(b''):
            if CTOK2 == CTOK:
                self.setverdict("fail", "Token should be different")
        pl_5 = self.frame.coap["pl"]
        self.next_skip_ack()

        # Step 6
        if not self.match_coap("server", CoAP(type=Any(CoAPType("con"), "ack"),
                                              code=2.04,
                                              mid=CMID2,
                                              tok=CTOK2)):
            raise self.Stop()
        if self.match_coap("server", CoAP(pl=Not(b"")), None):
            self.match_coap("server", CoAP(
                opt=Opt(CoAPOptionContentFormat()),
            ), "fail")

        self.next_skip_ack(optional=True)

        # Part B

        # Step 9
        self.chain()

        self.match_coap("client", CoAP(type="con", code="get",
                                       opt=Opt(CoAPOptionUriPath("validate"))))
        CMID3 = self.frame.coap["mid"]
        CTOK3 = self.frame.coap["tok"]
        if CMID3 is Not(b''):
            if CMID3 == CMID or CMID3 == CMID2:
                self.setverdict("fail", "Message ID should be different")
        if CTOK3 is Not(b''):
            if CTOK3 == CTOK or CTOK3 == CTOK2:
                self.setverdict("fail", "Token should be different")

        self.next_skip_ack()

        # Step 10
        if not self.match_coap("server", CoAP(type=Any(CoAPType("con"), "ack"),
                                              code=2.05,
                                              mid=CMID3,
                                              tok=CTOK3,
                                              opt=Opt(CoAPOptionETag(Not(ETAG1))),
                                              pl=pl_5)):
            raise self.Stop()

        ETAG2 = self.frame.coap["opt"][CoAPOptionETag]["val"]
        pl_10 = self.frame.coap["pl"]

        self.next_skip_ack(optional=True)

        self.chain()

        # Step 13
        self.match_coap("client", CoAP(type="con", code="put",
                                       opt=Opt(
                                           CoAPOptionUriPath("validate"),
                                           CoAPOptionContentFormat(),
                                       ),
                                       pl=All(Not(b""), Not(pl_3), Not(pl_10))))
        CMID4 = self.frame.coap["mid"]
        CTOK4 = self.frame.coap["tok"]
        if CMID4 is Not(b''):
            if CMID4 == CMID or CMID4 == CMID2 or CMID4 == CMID3:
                self.setverdict("fail", "Message ID should be different")
        if CTOK4 is Not(b''):
            if CTOK4 == CTOK or CTOK4 == CTOK2 or CTOK4 == CTOK3:
                self.setverdict("fail", "Token should be different")
        pl_13 = self.frame.coap["pl"]

        self.next_skip_ack()

        # Step 14
        self.match_coap("server", CoAP(type=Any(CoAPType("con"), "ack"),
                                       code=2.04,
                                       mid=CMID4,
                                       tok=CTOK4, ))
        if self.match_coap("server", CoAP(pl=Not(b"")), None):
            self.match_coap("server", CoAP(
                opt=Opt(CoAPOptionContentFormat()),
            ), "fail")

        self.next_skip_ack(optional=True)

        self.chain()

        # Step 17
        self.match_coap("client", CoAP(type="con", code="put",
                                       opt=Opt(
                                           CoAPOptionContentFormat(),
                                           CoAPOptionUriPath("validate"),
                                           CoAPOptionIfMatch(ETAG2),
                                       ),
                                       pl=All(Not(b""), Not(pl_13))))
        CMID5 = self.frame.coap["mid"]
        CTOK5 = self.frame.coap["tok"]
        if CMID5 is Not(b''):
            if CMID5 == CMID or CMID5 == CMID2 or CMID5 == CMID3 or CMID5 == CMID4:
                self.setverdict("fail", "Message ID should be different")
        if CTOK5 is Not(b''):
            if CTOK5 == CTOK or CTOK5 == CTOK2 or CTOK5 == CTOK3 or CTOK5 == CTOK4:
                self.setverdict("fail", "Token should be different")

        self.next_skip_ack()
        # Step 18
        self.match_coap("server", CoAP(type=Any(CoAPType("con"), "ack"),
                                       code=4.12,
                                       mid=CMID5,
                                       tok=CTOK5, ))
